# colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#ca9a96', '#facaa9']

import matplotlib.pyplot as plt
import numpy as np

data = {
    'MS/MS\n(DDA)': [567, 0, 0],
    'MS1\n(DDA)': [64, 242, 0],
    'MS1\n(0 eV)': [80, 247, 157],
    'MS1\n(10 eV)': [98, 249, 164],
    'MS1\n(20 eV)': [138, 261, 205]
}

categories = ['MS/MS collected in DDA',
              'No MS/MS collected in DDA',
              'New annotations from newly detected features']

colors = ['#56648a', '#8ca5c0', '#ca9a96']

# Set up the plot
bar_width = 0.65  # Width of each bar (0-1)
starting_width = 0.5  # Space between y-axis and first bar

# Create the stacked bar plot
fig, ax = plt.subplots(figsize=(4.5, 3))  # Increased height to accommodate top legend

x = np.arange(len(data))  # the label locations
bottom = np.zeros(len(data))

for i, cat in enumerate(categories):
    values = [data[key][i] for key in data.keys()]
    ax.bar(x + starting_width, values, bar_width, bottom=bottom, label=cat, color=colors[i], alpha=0.5)

    # Add individual numbers inside each bar
    for j, v in enumerate(values):
        if v > 0:
            ax.text(j + starting_width, bottom[j] + v / 2, str(v), ha='center', va='center',
                    fontsize=10.5, fontname='Arial', color='0.1')

    bottom += values

# Customize the plot
ax.set_ylabel('Annotated features', fontsize=15, fontname='Arial', labelpad=8, color='0.2')
ax.set_xticks(x + starting_width)
ax.set_xticklabels(data.keys())

# Customize ticks and remove top and right spines
ax.tick_params(axis='x', which='major', labelsize=12, pad=5, length=2, colors='0.2')
ax.tick_params(axis='y', which='major', labelsize=10, pad=1.5, length=2, colors='0.2')
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['bottom'].set_color('0.5')
ax.spines['left'].set_color('0.5')

# Set x-axis and y-axis label font to Arial and color to 0.2
for label in ax.get_xticklabels() + ax.get_yticklabels():
    label.set_fontname('Arial')
    label.set_color('0.2')

# # Customize legend
# legend = ax.legend(loc='upper center', bbox_to_anchor=(0.5, 1.35),
#                    frameon=False, ncol=1,
#                    prop={'family': 'Arial', 'size': 11.25})
#
# # Explicitly set the legend text color
# for text in legend.get_texts():
#     text.set_color('0.2')

# Add total value labels on top of the bars
for i, mode in enumerate(data.keys()):
    total = sum(data[mode])
    ax.text(i + starting_width, total, str(total), ha='center', va='bottom', fontsize=10.5, fontname='Arial', color='0.1')

# Adjust layout and display the plot
plt.tight_layout()

# Save as svg
plt.savefig('data/bar_plot.svg', transparent=True, bbox_inches='tight')

plt.show()


# Description: This file contains classes that are used in the main pipeline.


class PseudoMS2:
    def __init__(self, t_mz, t_mz_idx, mz_ls, int_ls, idx_ls):
        self.t_mz = t_mz  # this PseudoMS1 spectrum is generated starting from this mz
        self.t_mz_idx = t_mz_idx  # idx of t_mz in the mz_values
        self.mzs = mz_ls
        self.intensities = int_ls
        self.indices = idx_ls  # indices of mzs, for later assign intensities
        self.annotated = False
        self.annotation_ls = []  # list of SpecAnnotation objects


class SpecAnnotation:
    def __init__(self, db_name, idx, score, matched_peak):
        self.db_name = db_name
        self.search_eng_matched_id = idx  # index of the matched spec in the search engine
        self.score = score
        self.matched_peak = matched_peak
        self.spectral_usage = None
        self.matched_spec = None  # the matched spectrum in the search engine
        self.db_id = None
        self.name = None
        self.mz = None  # mz in PseudoMS2
        self.precursor_mz = None  # precursor mz of matched spec
        self.precursor_type = None
        self.formula = None
        self.inchikey = None
        self.instrument_type = None
        self.collision_energy = None
        self.centroided_peaks = None

    def __str__(self):
        return f"name {self.name}, score {self.score}, matched_peak {self.matched_peak}, db_id {self.db_id}"



"""
Note: This script requires Numba to be installed. You can install it using:
pip install numba

The first run of this script might be slower due to Numba's compilation time,
but subsequent runs should be much faster.
"""

import os

import numpy as np
from numba import njit
from scipy.sparse import csr_matrix, save_npz
from scipy.signal import savgol_filter


@njit
def peak_peak_correlation_numba(scan_idx1, int_seq1, scan_idx2, int_seq2, roi_min_length=3):
    i, j = 0, 0
    sum_x, sum_y, sum_xy, sum_x2, sum_y2 = 0.0, 0.0, 0.0, 0.0, 0.0
    n = 0

    while i < len(scan_idx1) and j < len(scan_idx2):
        if scan_idx1[i] < scan_idx2[j]:
            i += 1
        elif scan_idx1[i] > scan_idx2[j]:
            j += 1
        else:
            x, y = int_seq1[i], int_seq2[j]
            sum_x += x
            sum_y += y
            sum_xy += x * y
            sum_x2 += x * x
            sum_y2 += y * y
            n += 1
            i += 1
            j += 1

    if n < roi_min_length:
        return 0.0

    numerator = n * sum_xy - sum_x * sum_y
    denominator = np.sqrt((n * sum_x2 - sum_x ** 2) * (n * sum_y2 - sum_y ** 2))

    return numerator / denominator if denominator != 0 else 1.0


@njit
def calc_ppc_numba(roi_ids, roi_rts, roi_scan_idx_seqs, roi_int_seqs, roi_lengths, rt_tol,
                   roi_min_length=3, min_ppc=0.8):
    n_rois = len(roi_ids)
    rows = []
    cols = []
    data = []

    for i in range(n_rois):
        start_i = np.sum(roi_lengths[:i])
        end_i = start_i + roi_lengths[i]
        for j in range(i + 1, n_rois):
            if abs(roi_rts[i] - roi_rts[j]) <= rt_tol:
                start_j = np.sum(roi_lengths[:j])
                end_j = start_j + roi_lengths[j]
                ppc = peak_peak_correlation_numba(
                    roi_scan_idx_seqs[start_i:end_i], roi_int_seqs[start_i:end_i],
                    roi_scan_idx_seqs[start_j:end_j], roi_int_seqs[start_j:end_j],
                    roi_min_length
                )
                if ppc >= min_ppc:
                    rows.append(i)
                    cols.append(j)
                    data.append(ppc)

    return np.array(rows), np.array(cols), np.array(data)


def calc_all_ppc(d, rt_tol=0.1, roi_min_length=3, min_ppc=0.8, save=True, path=None):
    """
    Calculate peak-peak correlation matrix for all ROIs in the dataset
    """
    rois = sorted(d.rois, key=lambda x: x.id)
    n_rois = len(rois)

    roi_ids = np.array([roi.id for roi in rois], dtype=np.int32)
    roi_rts = np.array([roi.rt for roi in rois], dtype=np.float64)
    roi_lengths = np.array([len(roi.scan_idx_seq) for roi in rois], dtype=np.int32)
    roi_scan_idx_seqs = np.concatenate([np.array(roi.scan_idx_seq, dtype=np.int32) for roi in rois])
    roi_int_seqs = np.concatenate([np.array(roi.int_seq, dtype=np.float64) for roi in rois])

    # smoothing
    # roi_int_seqs = np.concatenate([peak_smooth(np.array(roi.int_seq, dtype=np.float64)) for roi in rois])

    rows, cols, data = calc_ppc_numba(
        roi_ids, roi_rts, roi_scan_idx_seqs, roi_int_seqs, roi_lengths, rt_tol, roi_min_length, min_ppc
    )

    # Create sparse matrix
    ppc_matrix = csr_matrix((data, (rows, cols)), shape=(n_rois, n_rois), dtype=np.float64)
    ppc_matrix = ppc_matrix + ppc_matrix.T
    ppc_matrix.setdiag(1.0)

    if save and path:
        os.makedirs(os.path.dirname(path), exist_ok=True)
        save_npz(path, ppc_matrix)

    return ppc_matrix


def peak_smooth(int_array, window_length=5, poly_order=2):
    """
    Smoothing the intensity sequence using Savitzky-Golay filter.
    If the sequence is shorter than the window length, adjust the window size accordingly.

    Parameters:
    - int_array (array-like): The intensity sequence to smooth.
    - window_length (int): The length of the filter window (must be odd).
    - polyorder (int): The order of the polynomial used to fit the samples.

    Returns:
    - smoothed_seq (numpy array): Smoothed version of the input intensity sequence.
    """
    int_len = len(int_array)

    if int_len <= poly_order:  # poly_order must be larger than 1
        return int_array

    # Ensure window_length is at most the length of int_seq and make it odd if necessary
    if window_length > int_len:
        window_length = int_len if int_len % 2 == 1 else int_len - 1

    return savgol_filter(int_array, window_length, poly_order)



import pandas as pd
import joblib
import numpy as np


def prepare_qry_spec():
    # ['Name', 'Precursor_type', 'Instrument_type', 'Collision_energy',
    #        'Precursor_mz', 'Formula', 'MW', 'CAS', 'NIST_No', 'Notes',
    #        'Instrument', 'Ionization', 'Collision_gas', 'Sample_inlet',
    #        'Spectrum_type', 'Ion_mode', 'InChIKey', 'Fragment_No', 'MS2mz',
    #        'MS2int', 'Frag_formula', 'Loss_formula']

    df = joblib.load('/Users/shipei/Documents/projects/ms2/ms2_lib/nist20/nist20/nist20.joblib')

    df = pd.DataFrame(df)

    # drop cols of 'Frag_formula', 'Loss_formula'
    df.drop(['Frag_formula', 'Loss_formula'], axis=1, inplace=True)

    df = df[df['Spectrum_type'] == 'MS2']

    # convert 'InChIKey' to str
    df['InChIKey'] = df['InChIKey'].astype(str)

    # print(df['Collision_energy'].value_counts())

    df['cleaned_CE'], df['CE_mode'] = zip(*df['Collision_energy'].apply(_clean_CE))

    # sort by Instrument_type, Ion_mode, InChIKey, Name, Precursor_type, Collision_gas, cleaned_CE, CE_mode
    df = df.sort_values(['Instrument_type', 'Ion_mode', 'InChIKey', 'Name', 'Precursor_type', 'Collision_gas', 'cleaned_CE', 'CE_mode'])

    # deduplicate by Instrument_type, Ion_mode, InChIKey, Name, Precursor_type, Collision_gas, CE_mode, keep first
    df = df.drop_duplicates(subset=['Instrument_type', 'Ion_mode', 'InChIKey', 'Name', 'Precursor_type', 'Collision_gas', 'CE_mode'], keep='first')

    df['peaks'] = df.apply(lambda x: np.array([[float(y) for y in x['MS2mz'].split(';')],
                                               [float(y) for y in x['MS2int'].split(';')]]).T, axis=1)

    # filter
    mask_1 = (df['cleaned_CE'] < 3) & (df['CE_mode'] == 'ev')
    mask_2 = (df['cleaned_CE'] < 3) & (df['CE_mode'] == 'nce')
    df = df[mask_1 | mask_2]
    # df = df[mask_1]

    df = df[df['Precursor_type'].isin(['[M+H]+', '[M-H]-'])]

    df.to_pickle('low_energy_nist20.pkl')
    df.to_csv('low_energy_nist20.tsv', sep='\t', index=False)
    print(df.shape[0])


def _clean_CE(x):
    try:
        return float(x), 'ev'
    except:
        if 'NCE' in x:

            # return float(x.split('=')[1].split('%')[0]), 'nce'

            if 'eV' in x:
                return float(x.split(' ')[1].split('eV')[0]), 'ev'
            else:
                return float(x.split('=')[1].split('%')[0]), 'nce'


def read_mgf_to_df(library_mgf, write_pkl=False, out_path=None):
    """
    Generate a dataframe from mgf file
    """
    with open(library_mgf, 'r') as file:
        spectrum_list = []
        for line in file:
            # empty line
            _line = line.strip()
            if not _line:
                continue
            elif line.startswith('BEGIN IONS'):
                spectrum = {}
                # initialize spectrum
                mz_list = []
                intensity_list = []
            elif line.startswith('END IONS'):
                if len(mz_list) == 0:
                    continue
                spectrum['mz_ls'] = mz_list
                spectrum['intensity_ls'] = intensity_list
                spectrum_list.append(spectrum)
                continue
            else:
                # if line contains '=', it is a key-value pair
                if '=' in _line:
                    # split by first '='
                    key, value = _line.split('=', 1)
                    spectrum[key] = value
                else:
                    # if no '=', it is a spectrum pair
                    this_mz, this_int = _line.split()
                    try:
                        mz_list.append(float(this_mz))
                        intensity_list.append(float(this_int))
                    except:
                        continue

    df = pd.DataFrame(spectrum_list)

    if write_pkl:
        if out_path is None:
            out_path = library_mgf.split('/')[-1].split('.')[0] + '_df.pkl'
        df.to_pickle(out_path)

    return df


if __name__ == '__main__':

    prepare_qry_spec()  # prepare low energy nist20, 5358 spectra

    # prepare nist20, for smiles
    # read_mgf_to_df('/Users/shipei/Documents/projects/ms2/ms2_lib/nist20/nist20/NIST20/NIST20_HR.mgf',
    #                write_pkl=True, out_path='nist20_df.pkl')




# Description: This file contains classes that are used in the main pipeline.
import numpy as np


class Spectrum:
    def __init__(self, mz_ls, int_ls):
        self.mzs = mz_ls
        self.intensities = int_ls
        self.annotation_ls = []  # list of SpecAnnotation objects


class PseudoMS2:

    def __init__(self, t_mz, mz_ls, int_ls, roi_ids, file_name, rt, mz_tol):
        self.t_mz = t_mz  # this PseudoMS2 spectrum is generated starting from this mz
        self.mzs = mz_ls
        self.intensities = int_ls
        self.roi_ids = roi_ids
        self.file_name = file_name
        self.rt = rt
        self.annotated = False
        self.annotation_ls = []  # list of SpecAnnotation objects

        # find the index of the target m/z in the pseudo MS2 spectrum, with the closest m/z
        t_mz_idx = np.argmin(np.abs(np.array(self.mzs) - self.t_mz))
        if abs(mz_ls[t_mz_idx] - self.t_mz) <= mz_tol:
            self.t_mz_idx = t_mz_idx
            self.t_mz_intensity = self.intensities[t_mz_idx]
        else:
            self.t_mz_idx = None  # if the target m/z is not in the pseudo MS2 spectrum
            self.t_mz_intensity = 0.0


class SpecAnnotation:
    def __init__(self, pseudo_ms2_spec_mzs, pseudo_ms2_spec_intensities,
                 prec_mz, db_name, idx, score, matched_peak, spectral_usage, mz_tol):
        self.db_name = db_name
        self.search_eng_matched_id = idx  # index of the matched spec in the search engine
        self.score = score
        self.matched_peak = matched_peak
        self.spectral_usage = spectral_usage
        self.matched_spec = None  # the matched spectrum in the search engine
        self.db_id = None
        self.name = None
        self.precursor_mz = prec_mz
        self.precursor_type = None
        self.formula = None
        self.inchikey = None
        self.instrument_type = None
        self.collision_energy = None
        self.centroided_peaks = None

        # annotation in this pseudo MS2 spectrum
        t_mz_idx = np.argmin(np.abs(np.array(pseudo_ms2_spec_mzs) - self.precursor_mz))
        if abs(pseudo_ms2_spec_mzs[t_mz_idx] - self.precursor_mz) <= mz_tol:
            self.pseudo_ms2_precursor_mz_idx = t_mz_idx
            self.pseudo_ms2_precursor_intensity = pseudo_ms2_spec_intensities[t_mz_idx]
        else:
            self.pseudo_ms2_precursor_mz_idx = None
            self.pseudo_ms2_precursor_intensity = 0.0

    def __str__(self):
        return f"name {self.name}, score {self.score}, matched_peak {self.matched_peak}, db_id {self.db_id}"


class AnnotatedPseudoMS2:
    """
    An annotation with pseudo MS2 spectrum stored
    """
    def __init__(self, spec_annotation, mzs, intensities):
        self.annotation = spec_annotation
        self.pseudo_ms2_mzs = mzs
        self.pseudo_ms2_intensities = intensities


class AlignedMS1Annotation:
    """
    for exporting aligned feature table with MS1 IDs
    store all possible MS1 IDs for a feature
    """

    def __init__(self, idx):
        self.df_idx = idx  # index of the matched feature in the feature table
        self.annotated_pseudo_ms2_list = []  # list of AnnotatedPseudoMS2 objects
        self.selected_annotated_pseudo_ms2 = None


def find_ms_info(file_name):
    """
    Find the type of MS and ion mode from the raw file.
    """

    ms_type = 'tof'
    ion_mode = 'positive'
    centroid = False

    with open(file_name, 'r') as f:
        for i, line in enumerate(f):
            if 'orbitrap' in line.lower() or 'Q Exactive' in line.lower():
                ms_type = 'orbitrap'
            if 'negative' in line.lower() or 'polarity="-"' in line.lower():
                ion_mode = 'negative'
            if "centroid spectrum" in line.lower() or 'centroided="1"' in line.lower():
                centroid = True
            if i > 200:
                break

    return ms_type, ion_mode, centroid


"""
run on server
"""

import pandas as pd
import numpy as np
from tqdm import tqdm
import networkx as nx


def main(mode='hilic_pos'):
    # read feature table
    df = pd.read_csv(f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/output/{mode}_aligned_feature_table.tsv',
                     sep='\t', low_memory=False)

    with open(f'data/{mode}_pMS2.mgf', 'w') as f:
        for _, row in tqdm(df.iterrows(), total=len(df), desc="Writing MGF"):
            if pd.notnull(row['pseudo_ms2']):

                pms2 = row['pseudo_ms2'].split(';')
                pms2 = [x.split(' ') for x in pms2][:-1]
                pms2 = [[float(x) for x in y] for y in pms2]
                pms2 = np.array(pms2, dtype=np.float32)

                # remove ions larger than precursor
                pms2 = pms2[pms2[:, 0] < row['m/z'] + 0.1]

                mzs = pms2[:, 0]
                intensities = pms2[:, 1]

                f.write('BEGIN IONS\n')
                f.write(f'NAME={row["MS1_annotation"]}\n')
                f.write(f'PEPMASS={row["m/z"]}\n')
                f.write(f'RTINSECONDS={row["RT"] * 60}\n')
                f.write(f'SCANS={round(row["ID"])}\n')

                for mz, intensity in zip(mzs, intensities):
                    f.write(f'{mz} {intensity}\n')
                f.write('END IONS\n\n')


def demo(frag_ls, name='Tyr',
         mode='hilic_pos'):
    # read feature table
    df = pd.read_csv(f'/Users/shipei/Documents/projects/ms1_id/data/PR000639/output/{mode}_aligned_feature_table.tsv',
                     sep='\t', low_memory=False)
    df[name] = False

    with open(f'data/{mode}_pMS2_{name}.mgf', 'w') as f:
        for _, row in tqdm(df.iterrows(), total=len(df), desc="Writing MGF"):
            if pd.notnull(row['pseudo_ms2']):

                pms2 = row['pseudo_ms2'].split(';')
                pms2 = [x.split(' ') for x in pms2][:-1]
                pms2 = [[float(x) for x in y] for y in pms2]
                pms2 = np.array(pms2, dtype=np.float32)

                # remove ions larger than precursor
                pms2 = pms2[pms2[:, 0] < row['m/z'] + 0.1]

                mzs = pms2[:, 0]
                intensities = pms2[:, 1]

                # Check if all fragments are present with >5% intensity
                mzs_5percent = mzs[intensities > 0.05 * np.max(intensities)]
                if not all(any(np.isclose(mzs_5percent, frag, atol=0.01)) for frag in frag_ls):
                    continue  # Skip this row and move to the next one

                df.loc[_, name] = True

                f.write('BEGIN IONS\n')
                f.write(f'NAME={row["MS1_annotation"]}\n')
                f.write(f'PEPMASS={row["m/z"]}\n')
                f.write(f'RTINSECONDS={row["RT"] * 60}\n')
                f.write(f'SCANS={round(row["ID"])}\n')

                for mz, intensity in zip(mzs, intensities):
                    f.write(f'{mz} {intensity}\n')
                f.write('END IONS\n\n')

    df = df[df[name]].reset_index(drop=True)
    df.to_csv(f'data/{mode}_{name}.tsv', sep='\t', index=False)


from msbuddy import Msbuddy, MsbuddyConfig
from msbuddy.base import MetaFeature, Spectrum


def annotate_formula():
    # read feature table
    df = pd.read_csv(f'/data/PR000639/output/hilic_pos_aligned_feature_table.tsv',
                     sep='\t', low_memory=False)

    df['annotated_formula_rank1'] = ''
    df['annotated_formula_rank2'] = ''
    df['annotated_formula_rank3'] = ''

    # create a MsbuddyConfig object
    msb_config = MsbuddyConfig(ppm=True,
                               ms1_tol=15,
                               ms2_tol=30,
                               halogen=False,
                               timeout_secs=60)
    # instantiate a Msbuddy object
    engine = Msbuddy(msb_config)

    all_metafeatures = []
    for idx, row in tqdm(df.iterrows(), total=len(df)):
        if pd.notnull(row['pseudo_ms2']):
            pms2 = row['pseudo_ms2'].split(';')
            pms2 = [x.split(' ') for x in pms2][:-1]
            pms2 = [[float(x) for x in y] for y in pms2]
            pms2 = np.array(pms2, dtype=np.float32)

            # remove ions larger than precursor
            pms2 = pms2[pms2[:, 0] < row['m/z'] + 0.1]

            mz_array = pms2[:, 0]
            int_array = pms2[:, 1]

            # create a Spectrum object
            ms2_spec = Spectrum(mz_array=mz_array, int_array=int_array)

            # create a MetaFeature object
            metafeature = MetaFeature(identifier=int(row['ID']),  # unique identifier for the MetaFeature object
                                      mz=row['m/z'],  # precursor m/z
                                      rt=row['RT'],  # retention time, can be None if not available
                                      charge=1,  # precursor charge
                                      ms2=ms2_spec)
        else:
            metafeature = MetaFeature(identifier=int(row['ID']),  # unique identifier for the MetaFeature object
                                      mz=row['m/z'],
                                      rt=row['RT'],
                                      charge=1)

        all_metafeatures.append(metafeature)

    # add to the Msbuddy object
    engine.add_data(all_metafeatures)

    # run the annotation
    engine.annotate_formula()

    # get the annotated data
    results = engine.get_summary()

    for result in results:
        mask = df['ID'] == result['identifier']
        df.loc[mask, 'annotated_formula_rank1'] = result['formula_rank_1']
        df.loc[mask, 'annotated_formula_rank2'] = result['formula_rank_2']
        df.loc[mask, 'annotated_formula_rank3'] = result['formula_rank_3']

    df.to_csv(f'data/hilic_pos_annotated_formula.tsv', sep='\t', index=False)


def add_annotations_to_network_file():
    # G = nx.read_graphml('data/mn/hilic_pos.graphml')
    G = nx.read_graphml('data/mn/hilic_pos_Tyr.graphml')

    # read feature table
    ms1_df = pd.read_csv(
        'data/hilic_pos_annotated_formula.tsv',
        sep='\t', low_memory=False)

    stats_df = pd.read_csv(
        '/data/PR000639/hilic_pos_statistical_analysis_with_outlier_removal.tsv',
        sep='\t', low_memory=False)

    # Iterate through all nodes in the graph
    for node, data in G.nodes(data=True):
        mz = data['mz']
        rt = data['rt']

        # find in ms1_df
        ms1_row = ms1_df[
            (ms1_df['m/z'].between(mz - 0.005, mz + 0.005)) & (ms1_df['RT'].between(rt - 0.05, rt + 0.05))].copy()
        if len(ms1_row) > 0:
            # sort by RT difference
            ms1_row['RT_diff'] = np.abs(ms1_row['RT'] - rt)
            ms1_row = ms1_row.sort_values('RT_diff', ascending=True).iloc[0]

            # if annotated
            if not pd.isnull(ms1_row['MS1_annotation']):
                data['MS1_annotation'] = ms1_row['MS1_annotation']
                data['MS1_formula'] = ms1_row['MS1_formula']
                data['MS1_annotated'] = True
                data['MS1_similarity'] = ms1_row['MS1_similarity']
            else:
                data['MS1_annotation'] = ''
                data['MS1_formula'] = ''
                data['MS1_annotated'] = False
                data['MS1_similarity'] = 0.0

            data['ID'] = ms1_row['ID']
            data['annotated_formula_rank1'] = ms1_row['annotated_formula_rank1']
            data['annotated_formula_rank2'] = ms1_row['annotated_formula_rank2']
            data['annotated_formula_rank3'] = ms1_row['annotated_formula_rank3']

            # add stats
            data['nonIBD_vs_CD_p_value_corrected'] = stats_df.loc[ms1_row['ID'], 'nonIBD_vs_CD_p_value_corrected']
            data['nonIBD_vs_UC_p_value_corrected'] = stats_df.loc[ms1_row['ID'], 'nonIBD_vs_UC_p_value_corrected']
            data['min_p_value'] = min(data['nonIBD_vs_CD_p_value_corrected'], data['nonIBD_vs_UC_p_value_corrected'])

            if data['min_p_value'] < 0.001:
                data['significant'] = 0
            elif data['min_p_value'] < 0.01:
                data['significant'] = 1
            elif data['min_p_value'] < 0.05:
                data['significant'] = 2
            else:
                data['significant'] = 3

        else:
            data['MS1_annotation'] = ''
            data['MS1_formula'] = ''
            data['MS1_annotated'] = False
            data['MS1_similarity'] = 0.0

    # Save the modified graph to a new GraphML file
    # nx.write_graphml(G, 'data/mn/hilic_pos_annotations.graphml')
    nx.write_graphml(G, 'data/mn/hilic_pos_Tyr_annotations.graphml')


import matplotlib.pyplot as plt
import seaborn as sns


def box_plot(id_list):
    metadata = pd.read_csv('/data/PR000639/metadata/metadata.csv')
    metadata['Diagnosis'] = metadata['Diagnosis'].replace({'nonIBD': 'Non-IBD'})
    sampleid_to_group = dict(zip(metadata['local_sample_id'], metadata['Diagnosis']))

    # read feature table
    df = pd.read_csv(
        '/data/PR000639/hilic_pos_statistical_analysis.tsv',
        sep='\t', low_memory=False)

    # Set up the plot style
    plt.rcParams['font.family'] = 'Arial'

    # Define color scheme
    diagnosis_colors = {'Non-IBD': '#56648a', 'UC': '#8ca5c0', 'CD': '#facaa9'}

    for id_num in id_list:
        # Filter the dataframe for the current ID
        id_data = df[df['ID'] == id_num]

        print('nonIBD vs CD: ', id_data['nonIBD_vs_CD_p_value'].values[0])
        print('nonIBD vs UC: ', id_data['nonIBD_vs_UC_p_value'].values[0])

        # Prepare data for plotting
        plot_data = []
        for col in id_data.columns:
            if col in sampleid_to_group:
                group = sampleid_to_group[col]
                value = id_data[col].values[0]
                plot_data.append({'Group': group, 'Value': value})

        plot_df = pd.DataFrame(plot_data)

        # Print quantile and median values for all groups
        print(f"\nQuantile and Median values for ID {id_num}:")
        for group in ['Non-IBD', 'CD', 'UC']:
            group_data = plot_df[plot_df['Group'] == group]['Value']
            if not group_data.empty:
                quantiles = group_data.quantile([0.25, 0.5, 0.75])
                print(f"{group}:")
                print(f"  25th percentile: {quantiles[0.25]:.2f}")
                print(f"  Median: {quantiles[0.5]:.2f}")
                print(f"  75th percentile: {quantiles[0.75]:.2f}")
            else:
                print(f"{group}: No data available")

        # Create the box plot
        fig, ax = plt.subplots(figsize=(1.8, 1.8))

        # Define groups and their positions
        groups = ['Non-IBD', 'CD', 'UC']
        positions = {group: i for i, group in enumerate(groups)}

        # Plot scatter points
        for group in groups:
            group_data = plot_df[plot_df['Group'] == group]['Value']
            if not group_data.empty:
                x = np.random.normal(positions[group], 0.05, len(group_data))
                ax.scatter(x, group_data, alpha=0.9, s=2.5, color=diagnosis_colors[group], edgecolors='none', zorder=1)

        # Create box plot
        box_width = 0.5
        box_data = [plot_df[plot_df['Group'] == group]['Value'] for group in groups if
                    not plot_df[plot_df['Group'] == group].empty]
        box_labels = [group for group in groups if not plot_df[plot_df['Group'] == group].empty]

        box_plot = ax.boxplot(box_data, positions=[positions[label] for label in box_labels],
                              patch_artist=True, labels=box_labels,
                              showfliers=False, widths=box_width)

        # Customize box appearance
        for element in ['whiskers', 'fliers', 'means', 'medians', 'caps']:
            plt.setp(box_plot[element], color='0.5')

        for patch, label in zip(box_plot['boxes'], box_labels):
            patch.set_facecolor(diagnosis_colors[label])
            patch.set_edgecolor('none')
            patch.set_alpha(0.6)

        # Customize the plot
        ax.set_ylabel('Peak height', fontsize=10, fontname='Arial', labelpad=3.5, color='0.2')
        # ax.set_yscale('log')

        # Set y-axis limits
        ax.set_ylim(0, 7e6)
        # ax.set_yticks([0, 5e6, 1e7])

        # Customize ticks and remove top and right spines
        ax.tick_params(axis='x', which='major', labelsize=10, pad=4, length=1.2, colors='0.2')
        ax.tick_params(axis='y', which='major', labelsize=10, pad=2.5, length=2, colors='0.2')
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('0.5')
        ax.spines['left'].set_color('0.5')

        # Set x-axis and y-axis label font to Arial and color to 0.2
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontname('Arial')
            label.set_color('0.2')

        # plt.title(f'Box Plot for ID: {id_num}', fontsize=14, fontname='Arial', color='0.2')
        plt.tight_layout()

        # Save as svg
        plt.savefig(f'data/boxplot_ID_{id_num}.svg', transparent=True, bbox_inches='tight')
        plt.show()


if __name__ == '__main__':
    ########## make MGF file
    # main('hilic_pos')
    # ########
    # hilic pos, tyrosine-related
    # demo(frag_ls=[182.0830, 165.0560, 136.0770], name='Tyr', mode='hilic_pos')

    ########
    # annotate_formula()

    ########
    # add_annotations_to_network_file()

    ########
    # box_plot(id_list=[19042])
    # box_plot(id_list=[118, 280, 1808, 19042, 1630, 16867, 17379, 2909])

    box_plot(id_list=[6696])



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt


def load_data():

    df = pd.read_csv('filtered_statistical_analysis_manual.tsv', sep='\t', low_memory=False)

    df = df[df['selected']].reset_index(drop=True)

    # sort by show_name
    df = df.sort_values('show_name').reset_index(drop=True)

    print(f"Number of selected features: {df.shape[0]}")
    return df


def heatmap(df, colormap='coolwarm'):
    plt.rcParams['font.family'] = 'Arial'

    metadata = pd.read_csv('/data/PR000639/metadata/metadata.csv')

    # Define the custom colors for diagnoses and sex
    diagnosis_colors = {'nonIBD': '#56648a', 'UC': '#8ca5c0', 'CD': '#facaa9'}
    sex_colors = {'Male': 'lightblue', 'Female': 'pink'}

    # Create ordering for the diagnoses and sex
    diagnosis_order = ['nonIBD', 'CD', 'UC']
    sex_order = ['Male', 'Female']

    # Create a new column in metadata that combines diagnosis and sex
    metadata['group'] = metadata['Diagnosis'] + ' ' + metadata['sex']

    # Create a custom order for the combined groups
    group_order = [f"{d} {s}" for d in diagnosis_order for s in sex_order]

    # Select data columns (starting from the 52nd column)
    data_columns = df.columns[51:]

    # Extract the data to plot
    plot_data = df[data_columns]

    # Cap the top 1% intensities to the 99th percentile for each row
    # plot_data = plot_data.apply(lambda row: row.clip(upper=np.percentile(row, 99)))

    # Apply log transformation (adding a small constant to avoid log(0))
    epsilon = 1e-10
    plot_data_log = np.log(plot_data + epsilon)

    # Perform z-score normalization within each row after log transformation
    plot_data_normalized = plot_data_log.apply(lambda x: (x - x.mean()) / x.std(), axis=1)

    # Calculate total z-scores for each sample
    total_z_scores = plot_data_normalized.sum()

    # Add total z-scores to metadata
    metadata['total_z_score'] = metadata['local_sample_id'].map(total_z_scores)

    # Create a sorting key for the group order
    metadata['group_order'] = metadata['group'].map({g: i for i, g in enumerate(group_order)})

    # Sort the metadata based on group order and total z-score within each group
    metadata_sorted = metadata.sort_values(['group_order', 'total_z_score'], ascending=[True, False])

    # Get the sorted sample names
    sorted_samples = metadata_sorted['local_sample_id']

    # Reorder columns based on sorted samples
    plot_data_normalized = plot_data_normalized[sorted_samples]

    # Create the heatmap
    plt.figure(figsize=(20, 5))
    g = sns.clustermap(plot_data_normalized,
                       col_cluster=False,
                       row_cluster=False,
                       yticklabels=df['show_name'],
                       xticklabels=False,
                       cmap=colormap,
                       center=0,
                       linewidths=0,
                       figsize=(20, 5))

    # Move y-axis labels to the left side and make ticks shorter
    g.ax_heatmap.yaxis.tick_left()
    g.ax_heatmap.yaxis.set_label_position('left')
    g.ax_heatmap.tick_params(axis='y', length=2)

    # Add colored bars to indicate diagnosis and sex
    ax_col = g.ax_col_dendrogram
    ax_col.set_visible(True)

    # Add color bars
    for i, sample in enumerate(sorted_samples):
        diagnosis = metadata.loc[metadata['local_sample_id'] == sample, 'Diagnosis'].iloc[0]
        sex = metadata.loc[metadata['local_sample_id'] == sample, 'sex'].iloc[0]
        ax_col.add_patch(plt.Rectangle((i, 0.5), 1, 0.5, facecolor=diagnosis_colors[diagnosis], edgecolor='none'))
        ax_col.add_patch(plt.Rectangle((i, 0), 1, 0.5, facecolor=sex_colors[sex], edgecolor='none'))

    ax_col.set_xlim(0, len(sorted_samples))
    ax_col.set_ylim(0, 1)
    ax_col.axis('off')

    plt.tight_layout()

    # save
    plt.savefig('heatmap.svg', transparent=True)
    plt.show()


def heatmap_2(df, colormap='coolwarm'):
    """
    heatmap, without sex
    """
    plt.rcParams['font.family'] = 'Arial'

    metadata = pd.read_csv('/data/PR000639/metadata/metadata.csv')

    # Define the custom colors for diagnoses
    diagnosis_colors = {'nonIBD': '#56648a', 'UC': '#8ca5c0', 'CD': '#facaa9'}

    # Create ordering for the diagnoses
    diagnosis_order = ['nonIBD', 'CD', 'UC']

    # Select data columns (starting from the 52nd column)
    data_columns = df.columns[51:]

    # Extract the data to plot
    plot_data = df[data_columns]

    # Cap the top 1% intensities to the 99th percentile for each row
    # plot_data = plot_data.apply(lambda row: row.clip(upper=np.percentile(row, 99)))

    # Apply log transformation (adding a small constant to avoid log(0))
    epsilon = 1e-10
    plot_data_log = np.log(plot_data + epsilon)

    # Perform z-score normalization within each row after log transformation
    plot_data_normalized = plot_data_log.apply(lambda x: (x - x.mean()) / x.std(), axis=1)

    # Calculate total z-scores for each sample
    total_z_scores = plot_data_normalized.sum()

    # Add total z-scores to metadata
    metadata['total_z_score'] = metadata['local_sample_id'].map(total_z_scores)

    # Create a sorting key for the diagnosis order
    metadata['diagnosis_order'] = metadata['Diagnosis'].map({d: i for i, d in enumerate(diagnosis_order)})

    # Sort the metadata based on diagnosis order and total z-score within each diagnosis
    metadata_sorted = metadata.sort_values(['diagnosis_order', 'total_z_score'], ascending=[True, False])

    # Get the sorted sample names
    sorted_samples = metadata_sorted['local_sample_id']

    # Reorder columns based on sorted samples
    plot_data_normalized = plot_data_normalized[sorted_samples]

    # Create the heatmap
    plt.figure(figsize=(20, 5))
    g = sns.clustermap(plot_data_normalized,
                       col_cluster=False,
                       row_cluster=False,
                       yticklabels=df['show_name'],
                       xticklabels=False,
                       cmap=colormap,
                       center=0,
                       linewidths=0,
                       figsize=(20, 5))

    # Move y-axis labels to the left side and make ticks shorter
    g.ax_heatmap.yaxis.tick_left()
    g.ax_heatmap.yaxis.set_label_position('left')
    g.ax_heatmap.tick_params(axis='y', length=2)

    # Add colored bars to indicate diagnosis
    ax_col = g.ax_col_dendrogram
    ax_col.set_visible(True)

    # Add color bars
    for i, sample in enumerate(sorted_samples):
        diagnosis = metadata.loc[metadata['local_sample_id'] == sample, 'Diagnosis'].iloc[0]
        ax_col.add_patch(plt.Rectangle((i, 0), 1, 1, facecolor=diagnosis_colors[diagnosis], edgecolor='none'))

    ax_col.set_xlim(0, len(sorted_samples))
    ax_col.set_ylim(0, 1)
    ax_col.axis('off')

    plt.tight_layout()

    # save
    plt.savefig('heatmap.svg', transparent=True)
    plt.show()


def bar_plot_1(df, fig_size=(2.2, 3.75)):
    max_fc = 2
    # Create the x_axis values
    df['x_axis'] = df['nonIBD_vs_CD_fold_change'].apply(lambda x: -x if x > 1 else 1 / x)
    df['x_axis'] = df['x_axis'].apply(lambda x: np.sign(x) * max_fc if abs(x) > max_fc else x)

    # Create the scatter plot
    fig, ax = plt.subplots(figsize=fig_size)

    # Set font to Arial
    plt.rcParams['font.family'] = 'Arial'

    # Plot lines connecting each point to the y-axis with breaks for large values
    for i, (x, p_value) in enumerate(zip(df['x_axis'], df['nonIBD_vs_CD_p_value_corrected'])):
        if abs(x) == max_fc:
            # Draw the main part of the line
            ax.plot([0, np.sign(x) * 1.2], [i, i], color='0.5', linewidth=1.8, zorder=1)
            # Draw the break
            break_length = 0.2
            ax.plot([np.sign(x) * (1.2 + break_length), np.sign(x) * max_fc], [i, i], color='0.5', linewidth=1.8, zorder=1)

            # Add double slash to indicate break
            slash_width = 0.15
            slash_height = 0.25
            x_start = np.sign(x) * 1.2
            y_start = i - slash_height / 2
            for j in range(2):
                ax.plot([x_start + j * slash_width * 0.7, x_start + slash_width + j * slash_width * 0.7],
                        [y_start, y_start + slash_height],
                        color='0.5', linewidth=1.5, zorder=2)
        else:
            ax.plot([0, x], [i, i], color='0.5', linewidth=1.8, zorder=1)

        # Add stars for p-value
        if p_value < 0.001:
            stars = '***'
        elif p_value < 0.01:
            stars = '**'
        elif p_value < 0.05:
            stars = '*'
        else:
            stars = ''

        if stars:
            x_text = max(x, 0) + 0.1 if x >= 0 else min(x, 0) - 0.1
            ax.text(x_text, i, stars, ha='left' if x >= 0 else 'right', va='center', fontsize=10)

    # Plot the scatter points
    ax.scatter(df['x_axis'], range(len(df)), color='orange', zorder=3)

    # Add dashed lines to separate metabolites
    for i in range(1, len(df)):
        ax.axhline(y=i - 0.5, color='gray', linestyle='--', linewidth=0.5, zorder=0)

    # Remove x-axis label and y-axis ticks
    ax.set_xlabel('')
    ax.set_yticks([])

    # Remove left and right spines, keep top and bottom
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    # Add a line at x=0
    ax.axvline(x=0, color='black', linewidth=0.5, zorder=2)

    # Set x-axis limits
    max_x = 2.3
    ax.set_xlim(-max_x, max_x)

    # Set x-ticks and labels
    xticks = [-2, -1, 0, 1, 2]
    ax.set_xticks(xticks)
    ax.set_xticklabels([f'{abs(int(x))}' for x in xticks])

    # Adjust the length of x-axis ticks
    ax.tick_params(axis='x', which='major', length=2, labelsize=8, pad=3)
    ax.tick_params(axis='x', which='minor', length=2)

    # Add background color for areas
    significance_threshold = 1
    ax.axvspan(-max_x, -significance_threshold, facecolor='lightgray', alpha=0.3, zorder=0)
    ax.axvspan(significance_threshold, max_x, facecolor='lightgray', alpha=0.3, zorder=0)

    # Adjust y-axis limits
    ax.set_ylim(-0.5, len(df) - 0.5)

    # Show the plot
    plt.tight_layout()

    # save
    plt.savefig('bar_plot_1.svg', transparent=True)
    plt.show()


def bar_plot_2(df, fig_size=(2.2, 3.75)):
    # Create the x_axis values
    df['x_axis'] = df['nonIBD_vs_UC_fold_change'].apply(lambda x: -x if x > 1 else 1 / x)
    df['x_axis'] = df['x_axis'].apply(lambda x: np.sign(x) * 3 if abs(x) > 3 else x)

    # Create the scatter plot
    fig, ax = plt.subplots(figsize=fig_size)

    # Set font to Arial
    plt.rcParams['font.family'] = 'Arial'

    # Plot lines connecting each point to the y-axis with breaks for large values
    for i, (x, p_value) in enumerate(zip(df['x_axis'], df['nonIBD_vs_UC_p_value_corrected'])):
        if abs(x) == 3:
            # Draw the main part of the line
            ax.plot([0, np.sign(x) * 2.0], [i, i], color='0.5', linewidth=1.8, zorder=1)
            # Draw the break
            break_length = 0.2
            ax.plot([np.sign(x) * (2.0 + break_length), np.sign(x) * 3], [i, i], color='0.5', linewidth=1.8, zorder=1)

            # Add double slash to indicate break
            slash_width = 0.15
            slash_height = 0.25
            x_start = np.sign(x) * 2.0
            y_start = i - slash_height / 2
            for j in range(2):
                ax.plot([x_start + j * slash_width * 0.7, x_start + slash_width + j * slash_width * 0.7],
                        [y_start, y_start + slash_height],
                        color='0.5', linewidth=1.5, zorder=2)
        else:
            ax.plot([0, x], [i, i], color='0.5', linewidth=1.8, zorder=1)

        # Add stars for p-value
        if p_value < 0.001:
            stars = '***'
        elif p_value < 0.01:
            stars = '**'
        elif p_value < 0.05:
            stars = '*'
        else:
            stars = ''

        if stars:
            x_text = max(x, 0) + 0.1 if x >= 0 else min(x, 0) - 0.1
            ax.text(x_text, i, stars, ha='left' if x >= 0 else 'right', va='center', fontsize=10)

    # Plot the scatter points
    ax.scatter(df['x_axis'], range(len(df)), color='orange', zorder=3)

    # Add dashed lines to separate metabolites
    for i in range(1, len(df)):
        ax.axhline(y=i - 0.5, color='gray', linestyle='--', linewidth=0.5, zorder=0)

    # Remove x-axis label and y-axis ticks
    ax.set_xlabel('')
    ax.set_yticks([])

    # Remove left and right spines, keep top and bottom
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    # Add a line at x=0
    ax.axvline(x=0, color='black', linewidth=0.5, zorder=2)

    # Set x-axis limits
    max_x = 2.3
    ax.set_xlim(-max_x, max_x)

    # Set x-ticks and labels
    xticks = [-2, -1, 0, 1, 2]
    ax.set_xticks(xticks)
    ax.set_xticklabels([f'{abs(int(x))}' for x in xticks])

    # Adjust the length of x-axis ticks
    ax.tick_params(axis='x', which='major', length=2, labelsize=8, pad=3)
    ax.tick_params(axis='x', which='minor', length=2)

    # Add background color for areas
    significance_threshold = 1
    ax.axvspan(-max_x, -significance_threshold, facecolor='lightgray', alpha=0.3, zorder=0)
    ax.axvspan(significance_threshold, max_x, facecolor='lightgray', alpha=0.3, zorder=0)

    # Adjust y-axis limits
    ax.set_ylim(-0.5, len(df) - 0.5)

    # Show the plot
    plt.tight_layout()

    # save
    plt.savefig('bar_plot_2.svg', transparent=True)
    plt.show()


if __name__ == '__main__':
    df = load_data()

    # heatmap(df, colormap='coolwarm')
    ### heatmap_2(df, colormap='coolwarm')

    bar_plot_1(df)
    bar_plot_2(df)



import pandas as pd
from tqdm import tqdm

import json


def load_workbench_df():

    with open('/data/PR000639/metadata/workbench_db/all', 'r') as file:
        data = json.load(file)

    # convert the data to a list of dictionaries
    rows = [value for key, value in data.items()]

    df = pd.DataFrame(rows)

    print(df.head())

    df.to_csv('data/workbench_db.tsv', sep='\t', index=False)


def add_inchikey_to_workbench_results():

    df = pd.read_csv('/data/PR000639/metadata/workbench_identifications.tsv',
                     sep='\t', low_memory=False)

    meta = pd.read_csv('data/workbench_db.tsv', sep='\t', low_memory=False)

    # dict from name to inchikey
    name_to_inchikey = dict(zip(meta['name'], meta['inchi_key']))

    # add inchikey to df
    df['inchikey'] = df['refmet_name'].map(name_to_inchikey)

    df.to_csv('data/workbench_identifications_with_inchikey.tsv', sep='\t', index=False)


if __name__ == '__main__':
    load_workbench_df()

    add_inchikey_to_workbench_results()



import numpy as np
import pyimzml.ImzMLParser as imzml
import matplotlib.pyplot as plt
from scipy.ndimage import median_filter
import matplotlib.colors as colors
from matplotlib.backends.backend_pdf import PdfPages
import pandas as pd
from tqdm import tqdm

def get_ion_image_from_parser(parser, target_mz, tolerance=0.01, intensity_threshold=0):
    max_x = max(x for x, y, z in parser.coordinates)
    max_y = max(y for x, y, z in parser.coordinates)
    image = np.zeros((max_y + 1, max_x + 1))

    for idx, (x, y, z) in enumerate(parser.coordinates):
        mzs, ints = parser.getspectrum(idx)
        mask = np.abs(mzs - target_mz) <= tolerance
        matched_intensities = ints[mask]

        if len(matched_intensities) > 0:
            intensity = np.max(matched_intensities)
            if intensity > intensity_threshold:
                image[y, x] = intensity

    return image

def plot_ion_image(ax, image, title, colormap='viridis', norm_max_intensity=None,
                   hot_spot_removal=False, hot_spot_percentile=99, median_filter_flag=False):
    if hot_spot_removal:
        percentile_top = np.percentile(image, hot_spot_percentile)
        image[image > percentile_top] = percentile_top

    if median_filter_flag:
        image = median_filter(image, size=3)

    if norm_max_intensity is None:
        norm_max_intensity = np.max(image)

    original_cmap = plt.get_cmap(colormap)
    colors_array = original_cmap(np.linspace(0.10, 1, 256))
    custom_cmap = colors.LinearSegmentedColormap.from_list("custom_" + colormap, colors_array)

    im = ax.imshow(image, cmap=custom_cmap, vmin=0, vmax=norm_max_intensity)
    ax.set_title(title, fontsize=8, wrap=True)
    ax.axis('off')
    return im

def plot_all_ion_images(parser, df, output_pdf):

    # df = df[(df['matched_score'] > 0.8) & (df['spectral_usage'] > 0.1)].reset_index(drop=True)

    total_metabolites = len(df)
    with PdfPages(output_pdf) as pdf:
        for i in tqdm(range(0, total_metabolites, 10), desc="Processing metabolites", unit="page"):
            fig, axs = plt.subplots(5, 2, figsize=(15, 30))
            axs = axs.ravel()

            for j in range(10):
                if i + j < total_metabolites:
                    row = df.iloc[i + j]
                    image = get_ion_image_from_parser(parser, row['precursor_mz'], tolerance=0.005)

                    title = f"{row['name']}\n{row['precursor_mz']}\n{row['precursor_type']}\nScore: {row['matched_score']:.2f}\nPeak: {row['matched_peak']:.4f}\nUsage: {row['spectral_usage']:.2f}"

                    im = plot_ion_image(axs[j], image, title,
                                        norm_max_intensity=None,
                                        hot_spot_removal=True)

                    # plt.colorbar(im, ax=axs[j], fraction=0.046, pad=0.04)
                else:
                    axs[j].axis('off')

            plt.tight_layout()
            pdf.savefig(fig)
            plt.close()

# Usage
imzml_file = "/imaging/mouse_liver/Mouse liver_DMAN_200x200_25um.imzML"
df = pd.read_csv("/imaging/mouse_liver/Mouse liver_DMAN_200x200_25um/ms1_id_annotations_derep.tsv", sep='\t')
output_pdf = "ion_images.pdf"

# Load the ImzML parser once
parser = imzml.ImzMLParser(imzml_file)

# Pass the parser to the function instead of the file path
plot_all_ion_images(parser, df, output_pdf)


import logging
from typing import Tuple

import numpy as np
from matchms.similarity.BaseSimilarity import BaseSimilarity
from matchms.similarity.spectrum_similarity_functions import find_matches
from matchms.typing import SpectrumType

logger = logging.getLogger("matchms")


class ReverseCosine(BaseSimilarity):
    """
    Calculate 'modified reverse cosine score' between mass spectra.
    reference spectrum goes first, query spectrum goes second.
    """
    # Set key characteristics as class attributes
    is_commutative = False

    # Set output data type, e.g. ("score", "float") or [("score", "float"), ("matches", "int")]
    # score_datatype = [("score", np.float64), ("matches", "int")]

    def __init__(self, tolerance: float = 0.1, mz_power: float = 0.0,
                 intensity_power: float = 1.0):
        """
        Parameters
        ----------
        tolerance:
            Peaks will be considered a match when <= tolerance apart. Default is 0.1.
        mz_power:
            The power to raise mz to in the cosine function. The default is 0, in which
            case the peak intensity products will not depend on the m/z ratios.
        intensity_power:
            The power to raise intensity to in the cosine function. The default is 1.
        """
        self.tolerance = tolerance
        self.mz_power = mz_power
        self.intensity_power = intensity_power

    def pair(self, reference: SpectrumType, query: SpectrumType):
        """Calculate modified cosine score between two spectra.

        Parameters
        ----------
        reference
            Single reference spectrum.
        query
            Single query spectrum.

        Returns
        -------

        Tuple with cosine score and number of matched peaks.
        """

        def get_matching_pairs():
            """Find all pairs of peaks that match within the given tolerance."""
            matching_pairs = collect_peak_pairs_reverse(spec1, spec2, self.tolerance, shift=0.0,
                                                        mz_power=self.mz_power,
                                                        intensity_power=self.intensity_power)
            if matching_pairs is None:
                matching_pairs = np.zeros((0, 4))
            if matching_pairs.shape[0] > 0:
                matching_pairs = matching_pairs[np.argsort(matching_pairs[:, 2])[::-1], :]
            return matching_pairs

        spec1 = reference.peaks.to_numpy
        spec2 = query.peaks.to_numpy

        matching_pairs = get_matching_pairs()

        if matching_pairs.shape[0] == 0:
            return 0., 0, 0., 0., [], []

        return score_best_matches_reverse(matching_pairs, spec1, spec2,
                                          self.mz_power, self.intensity_power)


def collect_peak_pairs_reverse(spec1: np.ndarray, spec2: np.ndarray,
                               tolerance: float, shift: float = 0, mz_power: float = 0.0,
                               intensity_power: float = 1.0):
    # pylint: disable=too-many-arguments
    """Find matching pairs between two spectra.

    Args
    ----
    spec1:
        Spectrum peaks and intensities as numpy array.
    spec2:
        Spectrum peaks and intensities as numpy array.
    tolerance
        Peaks will be considered a match when <= tolerance appart.
    shift
        Shift spectra peaks by shift. The default is 0.
    mz_power:
        The power to raise mz to in the cosine function. The default is 0, in which
        case the peak intensity products will not depend on the m/z ratios.
    intensity_power:
        The power to raise intensity to in the cosine function. The default is 1.

    Returns
    -------
    matching_pairs : numpy array
        Array of found matching peaks.
    """
    matches = find_matches(spec1[:, 0], spec2[:, 0], tolerance, shift)
    idx1 = [x[0] for x in matches]
    idx2 = [x[1] for x in matches]
    if len(idx1) == 0:
        return None
    matching_pairs = []
    for i, idx in enumerate(idx1):
        power_prod_spec1 = (spec1[idx, 0] ** mz_power) * (spec1[idx, 1] ** intensity_power)
        power_prod_spec2 = (spec2[idx2[i], 0] ** mz_power) * (spec2[idx2[i], 1] ** intensity_power)
        matching_pairs.append([idx, idx2[i], power_prod_spec1 * power_prod_spec2, power_prod_spec2])

    return np.array(matching_pairs.copy())  # idx1, idx2, int_spec1 * int_spec2, int_spec2


def score_best_matches_reverse(matching_pairs: np.ndarray, spec1: np.ndarray,
                               spec2: np.ndarray, mz_power: float = 0.0,
                               intensity_power: float = 1.0):
    """Calculate cosine-like score by multiplying matches. Does require a sorted
    list of matching peaks (sorted by intensity product)."""
    score = float(0.0)
    used_matches = int(0)
    used1 = set()
    used2 = set()
    spec2_matched_mz = np.array([], dtype=np.float64)
    spec2_matched_intensity = np.array([], dtype=np.float64)
    for i in range(matching_pairs.shape[0]):
        if not int(matching_pairs[i, 0]) in used1 and not int(matching_pairs[i, 1]) in used2:
            score += matching_pairs[i, 2]
            used1.add(int(matching_pairs[i, 0]))  # Every peak can only be paired once
            used2.add(int(matching_pairs[i, 1]))  # Every peak can only be paired once
            used_matches += 1

            spec2_idx = int(matching_pairs[i, 1])
            spec2_matched_mz = np.append(spec2_matched_mz, spec2[spec2_idx, 0])
            spec2_matched_intensity = np.append(spec2_matched_intensity, spec2[spec2_idx, 1])

    # Normalize score:
    spec1_power = np.power(spec1[:, 0], mz_power) * np.power(spec1[:, 1], intensity_power)
    spec2_power = np.power(spec2_matched_mz, mz_power) * np.power(spec2_matched_intensity, intensity_power)

    spec1_norm = np.sqrt(np.sum(np.power(spec1_power, 2)))
    spec2_norm = np.sqrt(np.sum(np.power(spec2_power, 2)))
    norm_product = spec1_norm * spec2_norm
    score /= norm_product

    # spectral usage (sum of used peaks / sum of all peaks)
    spec_usage_1 = np.sum(spec1[list(used1), 1]) / np.sum(spec1[:, 1])
    spec_usage_2 = np.sum(spec2[list(used2), 1]) / np.sum(spec2[:, 1])

    return score, used_matches, spec_usage_1, spec_usage_2, list(used1), list(used2)


if __name__ == "__main__":
    import numpy as np
    from matchms import Spectrum

    spec_1 = Spectrum(mz=np.array([100., 150, 200, 201]),
                      intensities=np.array([0.7, 0.2, 0.1, 0.2]),
                      metadata={"precursor_mz": 200.0})
    spec_2 = Spectrum(mz=np.array([105., 150, 190, 200]),
                      intensities=np.array([0.4, 0.2, 0.1, 0.5]),
                      metadata={"precursor_mz": 205.0})

    reverse_cosine = ReverseCosine(tolerance=0.05)
    score, matched_peak, spec_usage_ref, spec_usage_qry, idx_ls_ref, idx_ls_qry = (
        reverse_cosine.pair(spec_1, spec_2))

    print(score, matched_peak, spec_usage_ref, spec_usage_qry, idx_ls_ref, idx_ls_qry)



import os
import pickle
import numpy as np
from ms_entropy import read_one_spectrum

from ms1_id.lcms.utils import SpecAnnotation
from ms1_id.search_utils.flash_cos import FlashCos
from ms1_id.lcms.centroid_data import centroid_spectrum_for_search


def prepare_ms2_lib(ms2db,
                    mz_tol: float = 0.05,
                    peak_scale_k: float = None,
                    peak_intensity_power: float = 0.5):
    """
    prepare ms2 db using MSP or MGF formatted database
    :param ms2db: path to the MSP formatted database
    :param mz_tol: mz tolerance
    :param peak_scale_k: peak scaling factor, None for no scaling
    :param peak_intensity_power: peak intensity power, 0.5 for square root
    :return: a pickle file
    """

    replace_keys = {'precursormz': 'precursor_mz',
                    'pepmass': 'precursor_mz',
                    'precursortype': 'precursor_type',
                    'ionmode': 'ion_mode',
                    'instrumenttype': 'instrument_type',
                    'collisionenergy': 'collision_energy',
                    'spectrumid': 'comment'}

    db = []
    for a in read_one_spectrum(ms2db):
        # replace keys if exist
        for k, v in replace_keys.items():
            if k in a:
                a[v] = a.pop(k)

        # convert precursor_mz to float
        try:
            a['precursor_mz'] = float(a['precursor_mz'])
        except:
            a['precursor_mz'] = 0.0

        # ion_mode, harmonize
        if 'ion_mode' in a:
            ion_mode = a['ion_mode'].lower().strip()
            if ion_mode in ['p', 'positive']:
                a['ion_mode'] = 'positive'
            elif ion_mode in ['n', 'negative']:
                a['ion_mode'] = 'negative'
            else:
                a['ion_mode'] = 'unknown'

        db.append(a)

    print('Number of spectra in the database:', len(db))

    print('Initializing search engine')
    search_engine = FlashCos(max_ms2_tolerance_in_da=mz_tol * 1.005,
                             mz_index_step=0.0001,
                             peak_scale_k=peak_scale_k,
                             peak_intensity_power=peak_intensity_power)
    print('Building index')
    search_engine.build_index(db,
                              max_indexed_mz=2000,
                              precursor_ions_removal_da=0.5,
                              noise_threshold=0.01,
                              min_ms2_difference_in_da=mz_tol * 2.02,
                              clean_spectra=True)

    if peak_scale_k is None:
        new_path = os.path.splitext(ms2db)[0] + '.pkl'
    else:
        new_path = os.path.splitext(ms2db)[0] + f'_k{peak_scale_k}.pkl'

    # save as pickle
    with open(new_path, 'wb') as file:
        pickle.dump(search_engine, file)

    print(f"Pickle file saved to: {new_path}")
    return search_engine


def ms1_id_annotation(ms1_spec_ls, library_ls, mz_tol=0.05,
                      score_cutoff=0.8, min_matched_peak=6, min_spec_usage=0.05,
                      ion_mode=None, refine=True,
                      max_prec_rel_int_in_other_ms2=0.05,
                      save=False, save_path=None):
    """
    Perform ms1 annotation
    :param ms1_spec_ls: a list of PseudoMS1-like object
    :param library_ls: list of path to the pickle file, indexed library
    :param mz_tol: mz tolerance in Da, for rev cos matching
    :param score_cutoff: for rev cos
    :param min_matched_peak: for rev cos
    :param min_spec_usage: for rev cos
    :param ion_mode: str, ion mode, can be None (default), 'positive', or 'negative'
    :param refine: bool, refine the results
    :param max_prec_rel_int_in_other_ms2: float, maximum precursor relative intensity in other MS2 spectrum
    :param save: bool, save the results
    :param save_path: str, save directory
    :return: PseudoMS2-like object
    """

    # perform revcos matching
    ms1_spec_ls = ms1_id_revcos_matching(ms1_spec_ls, library_ls, mz_tol=mz_tol,
                                         ion_mode=ion_mode,
                                         score_cutoff=score_cutoff,
                                         min_matched_peak=min_matched_peak,
                                         min_spec_usage=min_spec_usage)

    # refine the results, to avoid wrong annotations (ATP, ADP, AMP all annotated at the same RT)
    if refine:
        print('Refining MS1 ID results...')
        ms1_spec_ls = refine_ms1_id_results(ms1_spec_ls, mz_tol=mz_tol,
                                            max_prec_rel_int=max_prec_rel_int_in_other_ms2)

    if save and save_path is not None:
        with open(save_path, 'wb') as file:
            pickle.dump(ms1_spec_ls, file)

    return ms1_spec_ls


def ms1_id_revcos_matching(ms1_spec_ls, library_ls, mz_tol=0.02,
                           ion_mode=None, score_cutoff=0.7,
                           min_matched_peak=3, min_spec_usage=0.0):
    """
    Perform MS1 annotation using open search for the entire spectrum, with filters similar to identity search.

    :param ms1_spec_ls: a list of PseudoMS2-like objects
    :param library: path to the pickle file, indexed library
    :param mz_tol: m/z tolerance in Da, for open matching
    :param ion_mode: str, ion mode, can be None (default), 'positive', or 'negative'
    :param score_cutoff: minimum score for matching
    :param min_matched_peak: minimum number of matched peaks
    :param min_spec_usage: minimum spectral usage
    :return: List of updated PseudoMS2-like objects
    """
    mz_tol = min(mz_tol, 0.05)  # indexed library mz_tol is 0.05

    # centroid peaks
    for spec in ms1_spec_ls:
        peaks = list(zip(spec.mzs, spec.intensities))
        peaks = np.asarray(peaks, dtype=np.float32, order="C")
        peaks = centroid_spectrum_for_search(peaks, width_da=0.05 * 2.015)
        spec.centroided_peaks = peaks

    # if library_ls is a string, convert to list
    if isinstance(library_ls, str):
        library_ls = [library_ls]

    # Load the data
    for library in library_ls:
        with open(library, 'rb') as file:
            search_eng = pickle.load(file)
        db_name = os.path.basename(library)

        for spec in ms1_spec_ls:

            if len(spec.centroided_peaks) < min_matched_peak:
                continue

            matching_result = search_eng.search(
                precursor_mz=2000.00,  # unused, open search
                peaks=spec.centroided_peaks,
                ms1_tolerance_in_da=mz_tol,
                ms2_tolerance_in_da=mz_tol,
                method="open",
                precursor_ions_removal_da=0.5,
                noise_threshold=0.0,
                min_ms2_difference_in_da=mz_tol * 2.02,
                reverse=True
            )

            score_arr, matched_peak_arr, spec_usage_arr = matching_result['open_search']

            # filter by matching cutoffs
            v = np.where((score_arr >= score_cutoff) &
                         (matched_peak_arr >= min_matched_peak) &
                         (spec_usage_arr >= min_spec_usage))[0]

            all_matches = []
            for idx in v:
                matched = {k.lower(): v for k, v in search_eng[idx].items()}

                this_ion_mode = matched.get('ion_mode', '')
                if ion_mode is not None and ion_mode != this_ion_mode:
                    continue

                # precursor should be in the pseudo MS2 spectrum
                precursor_mz = matched.get('precursor_mz', 0)
                if not any(np.isclose(np.array(spec.mzs), precursor_mz, atol=mz_tol)):
                    continue

                all_matches.append((idx, score_arr[idx], matched_peak_arr[idx], spec_usage_arr[idx]))

            if all_matches:
                spec.annotated = True
                for idx, score, matched_peaks, spectral_usage in all_matches:
                    matched = {k.lower(): v for k, v in search_eng[idx].items()}

                    annotation = SpecAnnotation(spec.mzs, spec.intensities, matched.get('precursor_mz'),
                                                db_name, idx, score, matched_peaks, spectral_usage, mz_tol)

                    annotation.name = matched.get('name', '')
                    annotation.precursor_type = matched.get('precursor_type', None)
                    annotation.formula = matched.get('formula', None)
                    annotation.inchikey = matched.get('inchikey', None)
                    annotation.instrument_type = matched.get('instrument_type', None)
                    annotation.collision_energy = matched.get('collision_energy', None)
                    annotation.db_id = matched.get('comment', None)
                    annotation.matched_spec = matched.get('peaks', None)

                    spec.annotation_ls.append(annotation)

    return ms1_spec_ls


def refine_ms1_id_results(ms1_spec_ls, mz_tol=0.01, max_prec_rel_int=0.05):
    """
    Refine MS1 ID results within each pseudo MS2 spectrum using a cumulative public spectrum approach.

    :param ms1_spec_ls: List of PseudoMS2-like objects
    :param mz_tol: m/z tolerance for comparing precursor masses
    :param max_prec_rel_int: Maximum relative intensity threshold for precursor in public spectrum
    :return: Refined list of PseudoMS2-like objects
    """
    for spec in ms1_spec_ls:
        if spec.annotated and len(spec.annotation_ls) > 1:
            # Sort annotations by precursor m/z in descending order
            spec.annotation_ls.sort(key=lambda x: x.precursor_mz, reverse=True)

            public_mz = np.array([])  # Public spectrum, all matched peaks
            public_intensity = np.array([])
            to_keep = []

            for annotation in spec.annotation_ls:

                current_precursor_mz = annotation.precursor_mz

                # Check if precursor appears in public spectrum
                if public_mz.size > 0:
                    mz_diff = np.abs(public_mz - current_precursor_mz)
                    min_diff_idx = np.argmin(mz_diff)
                    if mz_diff[min_diff_idx] <= mz_tol and public_intensity[min_diff_idx] > max_prec_rel_int:
                        continue

                to_keep.append(annotation)

                # Add the reference spectrum to the public spectrum
                ref_spectrum = np.array(annotation.matched_spec)
                ref_mz = ref_spectrum[:, 0]
                ref_intensity = ref_spectrum[:, 1] / np.max(ref_spectrum[:, 1])

                if public_mz.size == 0:
                    public_mz = ref_mz
                    public_intensity = ref_intensity
                else:
                    # Add peaks
                    public_mz = np.concatenate([public_mz, ref_mz])
                    public_intensity = np.concatenate([public_intensity, ref_intensity])

            # Update annotations
            spec.annotation_ls = to_keep

    return ms1_spec_ls


if __name__ == "__main__":
    ######### prepare the search engine #########
    prepare_ms2_lib(ms2db='/Users/shipei/Documents/projects/ms1_id/data/gnps.msp',
                    mz_tol=0.05, peak_scale_k=None, peak_intensity_power=0.5)
    prepare_ms2_lib(ms2db='/Users/shipei/Documents/projects/ms1_id/data/gnps.msp',
                    mz_tol=0.05, peak_scale_k=10, peak_intensity_power=0.5)

    ######### load the search engine #########
    # with open('/Users/shipei/Documents/projects/ms1_id/data/gnps.pkl', 'rb') as file:
    #     search_eng = pickle.load(file)
    #
    # print(search_eng)



"""
annotate adduct for single files and aligned features
"""

import numpy as np
from masscube.feature_grouping import peak_peak_correlation


def annotate_adduct(d, mz_tol=0.01, rt_tol=0.05):
    """
    A function to annotate adducts for single files

    Parameters
    ----------------------------------------------------------
    d: MSData object
    """

    d.rois.sort(key=lambda x: x.mz)
    roi_to_label = np.ones(len(d.rois), dtype=bool)

    # isotopes, in-source fragments
    for idx, r in enumerate(d.rois):
        if r.is_isotope or r.is_in_source_fragment:
            roi_to_label[idx] = False

    if d.params.ion_mode.lower() == "positive":
        default_single_adduct = "[M+H]+"
        default_double_adduct = "[M+2H]2+"
    elif d.params.ion_mode.lower() == "negative":
        default_single_adduct = "[M-H]-"
        default_double_adduct = "[M-2H]2-"

    adduct_single_list, adduct_double_list = _pop_adduct_list(_adduct_pos, _adduct_neg, d.params.ion_mode.lower())

    # find adducts by assuming the current roi is the [M+H]+ ion in positive mode and [M-H]- ion in negative mode
    for idx, r in enumerate(d.rois):

        if not roi_to_label[idx]:
            continue

        mol_mass = _calc_exact_mol_mass(r.mz, r.charge_state, d.params.ion_mode.lower())

        _adduct_list = adduct_single_list if r.charge_state == 1 else adduct_double_list

        for i, adduct in enumerate(_adduct_list):
            m = (mol_mass * _adduct_list[i]['m'] + _adduct_list[i]['mass']) / _adduct_list[i]['charge']
            v = np.logical_and(np.abs(d.roi_mz_seq - m) < mz_tol, np.abs(d.roi_rt_seq - r.rt) < rt_tol)
            v = np.where(np.logical_and(v, roi_to_label))[0]

            if len(v) == 0:
                continue

            if len(v) > 1:
                # select the one with the lowest RT difference
                v = v[np.argmin(np.abs(d.roi_rt_seq[v] - r.rt))]
            else:
                v = v[0]

            if peak_peak_correlation(r, d.rois[v]) > d.params.ppr:
                roi_to_label[v] = False
                d.rois[v].adduct_type = adduct['name']
                d.rois[v].adduct_parent_roi_id = r.id
                r.adduct_child_roi_id.append(d.rois[v].id)

        if len(r.adduct_child_roi_id) > 0:
            roi_to_label[idx] = False
            r.adduct_type = default_single_adduct if r.charge_state == 1 else default_double_adduct

    for r in d.rois:
        if r.adduct_type is None:
            r.adduct_type = default_single_adduct if r.charge_state == 1 else default_double_adduct


def _calc_exact_mol_mass(mz, charge_state, ion_mode):
    if ion_mode == 'positive':
        if charge_state == 1:
            return mz - 1.00727645223
        else:
            return mz * 2.0 - 2.01455290446
    else:
        if charge_state == 1:
            return mz + 1.00727645223
        else:
            return mz * 2.0 + 2.01455290446


def _pop_adduct_list(_adduct_pos, _adduct_neg, ion_mode):
    if ion_mode == 'positive':
        return _adduct_pos[1:-5], _adduct_pos[-4:]  # Remove M+H, M+2H
    else:
        return _adduct_neg[1:-3], _adduct_neg[-2:]  # Remove M-H, M-2H


_adduct_pos = [
    {'name': '[M+H]+', 'm': 1, 'charge': 1, 'mass': 1.00727645223},
    {'name': '[M+Na]+', 'm': 1, 'charge': 1, 'mass': 22.989220702},
    {'name': '[M+K]+', 'm': 1, 'charge': 1, 'mass': 38.9631579064},
    {'name': '[M+NH4]+', 'm': 1, 'charge': 1, 'mass': 18.03382555335},
    # {'name': '[M-H+2Na]+', 'm': 1, 'charge': 1, 'mass': 44.97116495177},
    {'name': '[M+H-H2O]+', 'm': 1, 'charge': 1, 'mass': -17.0032882318},
    {'name': '[M+H-2H2O]+', 'm': 1, 'charge': 1, 'mass': -35.01385291583},
    {'name': '[M+H-3H2O]+', 'm': 1, 'charge': 1, 'mass': -53.02441759986},

    {'name': '[2M+H]+', 'm': 2, 'charge': 1, 'mass': 1.00727645223},
    {'name': '[2M+Na]+', 'm': 2, 'charge': 1, 'mass': 22.989220702},
    {'name': '[2M+K]+', 'm': 2, 'charge': 1, 'mass': 38.9631579064},
    {'name': '[2M+NH4]+', 'm': 2, 'charge': 1, 'mass': 18.03382555335},
    # {'name': '[2M-H+2Na]+', 'm': 2, 'charge': 1, 'mass': 44.97116495177},
    {'name': '[2M+H-H2O]+', 'm': 2, 'charge': 1, 'mass': -17.0032882318},
    # {'name': '[2M+H-2H2O]+', 'm': 2, 'charge': 1, 'mass': -35.01385291583},
    # {'name': '[2M+H-3H2O]+', 'm': 2, 'charge': 1, 'mass': -53.02441759986},

    {'name': '[M+2H]2+', 'm': 1, 'charge': 2, 'mass': 2.01455290446},
    # {'name': '[M+H+Na]2+', 'm': 1, 'charge': 2, 'mass': 23.99649715423},
    # {'name': '[M+H+NH4]2+', 'm': 1, 'charge': 2, 'mass': 19.04110200558},
    {'name': '[M+Ca]2+', 'm': 1, 'charge': 2, 'mass': 39.961493703},
    {'name': '[M+Fe]2+', 'm': 1, 'charge': 2, 'mass': 55.93383917}
]

_adduct_neg = [
    {'name': '[M-H]-', 'm': 1, 'charge': 1, 'mass': -1.00727645223},
    {'name': '[M+Cl]-', 'm': 1, 'charge': 1, 'mass': 34.968304102},
    {'name': '[M+Br]-', 'm': 1, 'charge': 1, 'mass': 78.91778902},
    {'name': '[M+FA]-', 'm': 1, 'charge': 1, 'mass': 44.99710569137},
    {'name': '[M+Ac]-', 'm': 1, 'charge': 1, 'mass': 59.01275575583},
    {'name': '[M-H-H2O]-', 'm': 1, 'charge': 1, 'mass': -19.01784113626},

    {'name': '[2M-H]-', 'm': 2, 'charge': 1, 'mass': -1.00727645223},
    # {'name': '[2M+Cl]-', 'm': 2, 'charge': 1, 'mass': 34.968304102},
    # {'name': '[2M+Br]-', 'm': 2, 'charge': 1, 'mass': 78.91778902},
    {'name': '[2M+FA]-', 'm': 2, 'charge': 1, 'mass': 44.99710569137},
    {'name': '[2M+Ac]-', 'm': 2, 'charge': 1, 'mass': 59.01275575583},
    {'name': '[2M-H-H2O]-', 'm': 2, 'charge': 1, 'mass': -19.01784113626},

    {'name': '[M-2H]2-', 'm': 1, 'charge': 2, 'mass': -2.01455290446},
    # {'name': '[M-H+Cl]2-', 'm': 1, 'charge': 2, 'mass': 33.96157622977},
    # {'name': '[M-H+Br]2-', 'm': 1, 'charge': 2, 'mass': 77.91106114777},
]



import pandas as pd
from tqdm import tqdm
import pickle
import numpy as np
from _centroid_data import centroid_spectrum_for_search
from flash_cos import FlashCos


def search_main(library, qry_df, out_name,
                mz_tol=0.05, score_cutoff=0.6, min_matched_peak=1, min_spec_usage=0.0):

    with open(library, 'rb') as file:
        search_eng = pickle.load(file)

    all_matches = []
    for _, row in tqdm(qry_df.iterrows(), total=qry_df.shape[0]):
        peaks = row['peaks']
        peaks = centroid_spectrum_for_search(peaks, width_da=mz_tol * 2.015)

        # in reality, low-intensity peaks are hard to be retrieved using correlation
        # peaks = peaks[peaks[:, 1] > max(peaks[:, 1]) * 0.01]

        matching_result = search_eng.search(
            precursor_mz=float(row['Precursor_mz']),
            peaks=peaks,
            ms1_tolerance_in_da=mz_tol,
            ms2_tolerance_in_da=mz_tol,
            method="identity",
            precursor_ions_removal_da=0.5,
            noise_threshold=0.0,
            min_ms2_difference_in_da=mz_tol * 2.02,
            reverse=True
        )

        score_arr, matched_peak_arr, spec_usage_arr = matching_result['identity_search']

        # filter by matching cutoffs
        v = np.where((score_arr >= score_cutoff) &
                     (matched_peak_arr >= min_matched_peak))[0]

        matches = []
        for idx in v:
            matched = {k.lower(): v for k, v in search_eng[idx].items()}

            matches.append({
                'score': score_arr[idx],
                'matched_peak': matched_peak_arr[idx],
                'spec_usage': spec_usage_arr[idx],
                'qry_db_id': row['NIST_No'],
                'qry_name': row['Name'],
                'qry_prec_mz': row['Precursor_mz'],
                'qry_inchikey': row['InChIKey'],
                'matched_inchikey': matched.get('inchikey', ''),
                'matched_db_id': matched.get('comment', ''),
                'matched_name': matched.get('name', '')
            })

        all_matches.extend(matches)

    all_df = pd.DataFrame(all_matches)
    all_df.to_csv(f"{out_name}_results.tsv", sep='\t', index=False)


def add_nist_qry_smiles():
    # metadata from NIST20
    nist20 = pd.read_pickle('/src/scaling_fdr/nist20_df.pkl')
    nist20['NIST_No'] = nist20['TITLE'].apply(lambda x: x.split('NIST')[1].strip())
    # dict from NIST_No to SMILES
    nist_no_to_smiles = nist20.set_index('NIST_No')['SMILES'].to_dict()

    for mode in ['k0', 'k10']:
        df = pd.read_csv(f'{mode}_results.tsv', sep='\t')

        df['qry_db_id'] = df['qry_db_id'].astype(str)
        df['qry_smiles'] = df['qry_db_id'].map(nist_no_to_smiles, na_action='ignore')

        df['qry_inchikey_14'] = df['qry_inchikey'].str[:14]
        df['matched_inchikey_14'] = df['matched_inchikey'].str[:14]

        df.to_csv(f'{mode}_results.tsv', sep='\t', index=False)


if __name__ == '__main__':
    qry_df = pd.read_pickle('low_energy_nist20.pkl')

    library = '/Users/shipei/Documents/projects/ms1_id/data/gnps.pkl'
    search_main(library, qry_df, 'k0')

    library = '/Users/shipei/Documents/projects/ms1_id/data/gnps_k10.pkl'
    search_main(library, qry_df, 'k10')

    add_nist_qry_smiles()



"""
This is to get all features with high PPCs, generate pseudo MS1 for a single file
"""
import os
import pickle
import numpy as np

from ms1_id.lcms.utils import PseudoMS2
from ms1_id.lcms.export import write_pseudoms2_to_mgf


def retrieve_pseudo_ms2_spectra(config):
    """
    Retrieve pseudo MS1 spectra for all files
    :param config: Config object
    :return: dictionary of pseudo MS2 spectra
    """
    pseudo_ms2_spectra = []

    files = os.listdir(config.single_file_dir)
    for file in files:
        if file.endswith('_pseudoMS2_annotated.pkl'):
            try:
                with open(os.path.join(config.single_file_dir, file), 'rb') as f:
                    new_pseudo_ms2_spectra = pickle.load(f)
                    pseudo_ms2_spectra.extend(new_pseudo_ms2_spectra)
            except:
                continue

    return pseudo_ms2_spectra


def generate_pseudo_ms2(msdata, ppc_matrix,
                        mz_tol=0.01,
                        min_ppc=0.8,
                        min_cluster_size=4,
                        roi_min_length=3,
                        roi_max_rt_range=2.0,
                        save_dir=None):
    """
    Generate pseudo MS2 spectra for a single file
    :param msdata: MSData object
    :param ppc_matrix: sparse matrix of PPC scores
    :param roi_min_length: minimum length of ROIs to consider for clustering
    :param roi_max_rt_range: maximum RT range of ROIs to consider for clustering
    :param mz_tol: m/z tolerance
    :param min_ppc: minimum PPC score for clustering
    :param min_cluster_size: minimum number of ROIs in a cluster
    :param save_dir: directory to save the pseudo MS2 spectra
    """

    pseudo_ms2_spectra = _perform_clustering(msdata, ppc_matrix,
                                             mz_tol=mz_tol,
                                             min_ppc=min_ppc,
                                             min_cluster_size=min_cluster_size,
                                             roi_min_length=roi_min_length,
                                             roi_max_rt_range=roi_max_rt_range)

    write_pseudoms2_to_mgf(pseudo_ms2_spectra, save_dir, msdata.file_name)

    return pseudo_ms2_spectra


def _perform_clustering(msdata, ppc_matrix, mz_tol=0.01, min_ppc=0.8, min_cluster_size=4,
                        roi_min_length=3, roi_max_rt_range=1.0):
    """
    Perform clustering on ROIs based on PPC scores and m/z values,
    considering only m/z values smaller than the target m/z for each ROI.
    """
    # Filter ROIs based on minimum length and isotope status using a generator expression
    valid_rois = (roi for roi in msdata.rois if roi.length >= roi_min_length
                  # and (max(roi.rt_seq) - min(roi.rt_seq) <= roi_max_rt_range)
                  and not roi.is_isotope)

    # Sort ROIs by m/z values and create a mapping
    sorted_rois = sorted(valid_rois, key=lambda roi: roi.mz)

    # Create a new PPC matrix with only valid ROIs
    valid_indices = [msdata.rois.index(roi) for roi in sorted_rois]
    new_ppc_matrix = ppc_matrix[valid_indices][:, valid_indices]

    pseudo_ms2_spectra = []

    for i, roi in enumerate(sorted_rois):
        t_mz = roi.mz  # Set the target m/z as the current ROI's m/z

        # Find all ROIs with PPC scores above the threshold
        cluster_indices = new_ppc_matrix[i].nonzero()[1]
        cluster_scores = new_ppc_matrix[i, cluster_indices].toarray().flatten()

        # cluster_indices = cluster_indices[(cluster_scores >= min_ppc) &
        #                                   (np.array([sorted_rois[idx].mz for idx in cluster_indices]) <= t_mz + 1e-2)]
        cluster_indices = cluster_indices[cluster_scores >= min_ppc]

        if len(cluster_indices) >= min_cluster_size:
            # Form a pseudo MS2 spectrum
            cluster_rois = [sorted_rois[idx] for idx in cluster_indices]

            mz_ls = [roi.mz for roi in cluster_rois]
            int_ls = [roi.peak_height for roi in cluster_rois]
            roi_ids = [roi.id for roi in cluster_rois]
            # rt = np.mean([roi.rt for roi in cluster_rois])
            rt = roi.rt

            pseudo_ms2_spectra.append((t_mz, mz_ls, int_ls, roi_ids, msdata.file_name, rt, mz_tol))

    # Convert to PseudoMS2 objects after all processing
    pseudo_ms2_spectra = [PseudoMS2(*spec) for spec in pseudo_ms2_spectra]

    return pseudo_ms2_spectra




import multiprocessing as mp
import os
import tempfile

import numpy as np
from numba import njit
from scipy.sparse import csr_matrix, save_npz, load_npz
from tqdm import tqdm


@njit
def _mz_correlation(intensities1, intensities2, min_overlap=5):
    # Remove zero intensities
    non_zero_mask = (intensities1 != 0) & (intensities2 != 0)
    x = intensities1[non_zero_mask]
    y = intensities2[non_zero_mask]

    n = len(x)
    if n < min_overlap:
        return 0.0

    sum_x = np.sum(x)
    sum_y = np.sum(y)
    sum_xy = np.sum(x * y)
    sum_x2 = np.sum(x * x)
    sum_y2 = np.sum(y * y)

    numerator = n * sum_xy - sum_x * sum_y
    denominator = np.sqrt((n * sum_x2 - sum_x ** 2) * (n * sum_y2 - sum_y ** 2))

    return numerator / denominator if denominator != 0 else 0.0


def worker(start_idx, end_idx, mmap_filename, intensity_matrix_shape, min_overlap,
           min_cor, return_dict):
    intensity_matrix = np.memmap(mmap_filename, dtype=np.float64, mode='r', shape=intensity_matrix_shape)
    n_mzs = intensity_matrix_shape[0]

    rows = []
    cols = []
    data = []

    for i in range(start_idx, end_idx):
        for j in range(i + 1, n_mzs):
            corr = _mz_correlation(intensity_matrix[i], intensity_matrix[j], min_overlap)
            if corr >= min_cor:
                rows.append(i)
                cols.append(j)
                data.append(corr)

    return_dict[start_idx] = (rows, cols, data)


def calc_all_mz_correlations(intensity_matrix, min_overlap=5, min_cor=0.8,
                             save=True, save_dir=None, n_processes=None, chunk_size=500):
    """
    Calculate m/z correlation matrix for MS imaging data using multiprocessing and numpy memmap

    :param intensity_matrix: 2D numpy array where rows are m/z values and columns are spectra
    :param min_overlap: Minimum number of overlapping spectra between two ions
    :param min_cor: Minimum correlation value to keep
    :param save: Boolean indicating whether to save the result
    :param save_dir: Directory to save the result if save is True
    :param n_processes: Number of processes to use (default: number of CPU cores)
    :param chunk_size: Number of rows to process in each chunk
    :return: Sparse correlation matrix
    """

    # check if result files exist
    if save_dir is not None:
        path = os.path.join(save_dir, 'mz_correlation_matrix.npz')
        if os.path.exists(path):
            print("Loading existing correlation matrix...")
            return load_npz(path)

    n_mzs, n_spectra = intensity_matrix.shape

    n_processes = n_processes or mp.cpu_count()

    # Create a temporary memmap file
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    mmap_filename = temp_file.name
    mmap_array = np.memmap(mmap_filename, dtype=np.float64, mode='w+', shape=intensity_matrix.shape)
    mmap_array[:] = intensity_matrix[:]
    mmap_array.flush()

    # Prepare chunks
    chunks = [(i, min(i + chunk_size, n_mzs)) for i in range(0, n_mzs, chunk_size)]

    print(f"Calculating m/z spatial correlations...")

    if n_processes == 1:
        # Non-parallel processing
        all_rows = []
        all_cols = []
        all_data = []
        for start, end in tqdm(chunks, desc="Processing chunks"):
            rows, cols, data = [], [], []
            for i in range(start, end):
                for j in range(i + 1, n_mzs):
                    corr = _mz_correlation(mmap_array[i], mmap_array[j], min_overlap)
                    if corr >= min_cor:
                        rows.append(i)
                        cols.append(j)
                        data.append(corr)
            all_rows.extend(rows)
            all_cols.extend(cols)
            all_data.extend(data)
    else:
        # Parallel processing
        manager = mp.Manager()
        return_dict = manager.dict()

        with mp.Pool(processes=n_processes) as pool:
            jobs = [
                pool.apply_async(worker, (start, end, mmap_filename, intensity_matrix.shape,
                                          min_overlap, min_cor, return_dict))
                for start, end in chunks
            ]

            for job in tqdm(jobs, desc="Processing chunks"):
                job.get()  # Wait for the job to complete

        all_rows = []
        all_cols = []
        all_data = []
        for result in return_dict.values():
            all_rows.extend(result[0])
            all_cols.extend(result[1])
            all_data.extend(result[2])

    corr_matrix = csr_matrix((all_data, (all_rows, all_cols)), shape=(n_mzs, n_mzs), dtype=np.float64)
    corr_matrix = corr_matrix + corr_matrix.T
    corr_matrix.setdiag(1.0)

    if save and save_dir:
        path = os.path.join(save_dir, 'mz_correlation_matrix.npz')
        print(f"Saving correlation matrix to {path}...")
        save_npz(path, corr_matrix)

    # Clean up the temporary memmap file
    os.unlink(mmap_filename)

    return corr_matrix



import os
import pickle
from multiprocessing import Pool, cpu_count
from typing import List

import numpy as np
from tqdm import tqdm

from ms1_id.msi.centroid_data import centroid_spectrum_for_search
from ms1_id.msi.utils_imaging import SpecAnnotation
from ms1_id.search_utils.flash_cos import FlashCos


def validate_library_path(library_path):
    """
    Validate the library path
    :param library_path: str or list of str
    :return: list of str
    """
    # if library_ls is a string, convert to list
    if isinstance(library_path, str):
        library_path = [library_path]

    for path in library_path:
        if not os.path.exists(path):
            raise FileNotFoundError(f"Library file not found: {path}")

    return library_path


def ms1_id_annotation(ms1_spec_ls, library_ls, n_processes=None,
                      mz_tol=0.05,
                      score_cutoff=0.6, min_matched_peak=4, min_spec_usage=0.0,
                      ion_mode=None,
                      refine=False,
                      max_prec_rel_int_in_other_ms2=0.05,
                      save=False, save_dir=None,
                      chunk_size=1000):
    """
    Perform ms1 annotation
    :param ms1_spec_ls: a list of PseudoMS2-like object
    :param library: path to the pickle file, indexed library
    :param n_processes: number of processes to use
    :param mz_tol: mz tolerance in Da, for rev cos matching
    :param score_cutoff: for rev cos
    :param min_matched_peak: for rev cos
    :param peak_scale_k: for rev cos, peak scaling factor
    :param ion_mode: str, ion mode. If None, all ion modes are considered
    :param refine: bool, whether to refine the results
    :param max_prec_rel_int_in_other_ms2: float, maximum relative intensity of precursor in other MS2 spectra
    :param save: bool, whether to save the results
    :param save_dir: str, directory to save the results
    :param chunk_size: int, number of spectra to process in each parallel task
    :return: PseudoMS2-like object
    """

    # check if results are already annotated
    if save_dir:
        save_path = os.path.join(save_dir, 'pseudo_ms2_annotated.pkl')
        if os.path.exists(save_path):
            with open(save_path, 'rb') as file:
                ms1_spec_ls = pickle.load(file)
            return ms1_spec_ls

    if n_processes is None:
        n_processes = max(1, cpu_count() // 5)  # ms2 library is large, for RAM usage

    chunk_size = min(chunk_size, len(ms1_spec_ls))

    # Perform centroiding for all spectra before annotation
    ms1_spec_ls = centroid_all_spectra(ms1_spec_ls, n_processes)

    # perform revcos matching
    ms1_spec_ls = ms1_id_revcos_matching(ms1_spec_ls, library_ls, n_processes=n_processes,
                                         mz_tol=mz_tol,
                                         ion_mode=ion_mode,
                                         score_cutoff=score_cutoff,
                                         min_matched_peak=min_matched_peak,
                                         min_spec_usage=min_spec_usage,
                                         chunk_size=chunk_size)

    # refine the results, to avoid wrong annotations (ATP, ADP, AMP all annotated at the same RT)
    if refine:
        print('Refining MS1 ID results...')
        ms1_spec_ls = refine_ms1_id_results(ms1_spec_ls, mz_tol=mz_tol,
                                            max_prec_rel_int=max_prec_rel_int_in_other_ms2)

    if save:
        save_path = os.path.join(save_dir, 'pseudo_ms2_annotated.pkl')
        with open(save_path, 'wb') as file:
            pickle.dump(ms1_spec_ls, file)

    return ms1_spec_ls


def centroid_all_spectra(ms1_spec_ls, n_processes):
    """
    Centroid all spectra in parallel
    """
    with Pool(processes=n_processes) as pool:
        ms1_spec_ls = list(tqdm(pool.imap(_centroid_spectrum, ms1_spec_ls),
                                total=len(ms1_spec_ls),
                                desc="Centroiding spectra"))
    return ms1_spec_ls


def _centroid_spectrum(spec):
    peaks = list(zip(spec.mzs, spec.intensities))
    peaks = np.asarray(peaks, dtype=np.float32, order="C")
    spec.centroided_peaks = centroid_spectrum_for_search(peaks, width_da=0.05 * 2.015)
    return spec


def ms1_id_revcos_matching(ms1_spec_ls, library_ls, n_processes=None,
                           mz_tol=0.05,
                           ion_mode=None,
                           score_cutoff=0.7,
                           min_matched_peak=4,
                           min_spec_usage=0.0,
                           chunk_size=500) -> List:
    """
    Perform MS1 annotation using parallel open search for the entire spectrum, with filters similar to identity search.

    :param ms1_spec_ls: a list of PseudoMS2-like objects
    :param library_ls: path to the pickle file, indexed library
    :param n_processes: number of processes to use
    :param mz_tol: m/z tolerance in Da, for open matching
    :param ion_mode: str, ion mode
    :param score_cutoff: minimum score for matching
    :param min_matched_peak: minimum number of matched peaks
    :param min_spec_usage: minimum spectral usage
    :param chunk_size: number of spectra to process in each parallel task
    :return: List of updated PseudoMS2-like objects
    """
    mz_tol = min(mz_tol, 0.05)  # indexed library mz_tol is 0.05

    # Load all libraries
    search_engines = []
    for library in library_ls:
        with open(library, 'rb') as file:
            search_eng = pickle.load(file)
        db_name = os.path.basename(library)
        search_engines.append((search_eng, db_name))
        print(f"Loaded library: {db_name}")

    # Prepare chunks
    chunks = [ms1_spec_ls[i:i + chunk_size] for i in range(0, len(ms1_spec_ls), chunk_size)]

    # Prepare arguments for parallel processing
    args_list = [(chunk, search_engines, mz_tol, ion_mode, score_cutoff, min_matched_peak, min_spec_usage)
                 for chunk in chunks]

    # Use multiprocessing to process chunks in parallel
    with Pool(processes=n_processes) as pool:
        results = list(tqdm(pool.imap(_process_chunk_multi_lib, args_list), total=len(args_list), desc="Annotation: "))

    # Flatten results
    return [spec for chunk_result in results for spec in chunk_result]


def _process_chunk_multi_lib(args):
    chunk, search_engines, mz_tol, ion_mode, score_cutoff, min_matched_peak, min_spec_usage = args

    for spec in chunk:
        for search_eng, db_name in search_engines:

            if len(spec.centroided_peaks) < min_matched_peak:
                continue

            # open search
            matching_result = search_eng.search(
                precursor_mz=2000.00,  # unused, open search
                peaks=spec.centroided_peaks,
                ms1_tolerance_in_da=mz_tol,
                ms2_tolerance_in_da=mz_tol,
                method="open",
                precursor_ions_removal_da=0.5,
                noise_threshold=0.0,
                min_ms2_difference_in_da=mz_tol * 2.02,
                reverse=True
            )

            score_arr, matched_peak_arr, spec_usage_arr = matching_result['open_search']

            # filter by matching cutoffs
            v = np.where((score_arr >= score_cutoff) &
                         (matched_peak_arr >= min_matched_peak) &
                         (spec_usage_arr >= min_spec_usage))[0]

            all_matches = []
            nonzero_mask = np.array(spec.intensities) > 0
            nonzero_mzs = np.array(spec.mzs)[nonzero_mask]
            for idx in v:
                matched = {k.lower(): v for k, v in search_eng[idx].items()}

                this_ion_mode = matched.get('ion_mode', '')
                if ion_mode is not None and ion_mode != this_ion_mode:
                    continue

                # precursor should be in the pseudo MS2 spectrum with intensity > 0
                precursor_mz = matched.get('precursor_mz', 0)
                if not any(np.isclose(nonzero_mzs, precursor_mz, atol=mz_tol)):
                    continue

                all_matches.append((idx, score_arr[idx], matched_peak_arr[idx], spec_usage_arr[idx]))

            if all_matches:
                spec.annotated = True
                for idx, score, matched_peaks, spectral_usage in all_matches:
                    matched = {k.lower(): v for k, v in search_eng[idx].items()}

                    annotation = SpecAnnotation(db_name, idx, score, matched_peaks)
                    annotation.spectral_usage = spectral_usage

                    annotation.name = matched.get('name', '')

                    matched_spec_precursor_mz = matched.get('precursor_mz')
                    annotation.precursor_mz = matched_spec_precursor_mz
                    annotation.mz = min(nonzero_mzs, key=lambda x: abs(x - matched_spec_precursor_mz))

                    annotation.precursor_type = matched.get('precursor_type', None)
                    annotation.formula = matched.get('formula', None)
                    annotation.inchikey = matched.get('inchikey', None)
                    annotation.instrument_type = matched.get('instrument_type', None)
                    annotation.collision_energy = matched.get('collision_energy', None)
                    annotation.db_id = matched.get('comment', None)
                    annotation.matched_spec = matched.get('peaks', None)

                    spec.annotation_ls.append(annotation)

    return chunk


def refine_ms1_id_results(ms1_spec_ls, mz_tol=0.01, max_prec_rel_int=0.05):
    """
    Refine MS1 ID results within each pseudo MS2 spectrum using a cumulative public spectrum approach.

    :param ms1_spec_ls: List of PseudoMS2-like objects
    :param mz_tol: m/z tolerance for comparing precursor masses
    :param max_prec_rel_int: Maximum relative intensity threshold for precursor in public spectrum
    :return: Refined list of PseudoMS2-like objects
    """
    for spec in ms1_spec_ls:
        if spec.annotated and len(spec.annotation_ls) > 1:
            # Sort annotations by precursor m/z in descending order
            spec.annotation_ls.sort(key=lambda x: x.precursor_mz, reverse=True)

            public_mz = np.array([])  # Public spectrum, all matched peaks
            public_intensity = np.array([])
            to_keep = []

            for annotation in spec.annotation_ls:

                current_precursor_mz = annotation.precursor_mz

                # Check if precursor appears in public spectrum
                if public_mz.size > 0:
                    mz_diff = np.abs(public_mz - current_precursor_mz)
                    min_diff_idx = np.argmin(mz_diff)
                    if mz_diff[min_diff_idx] <= mz_tol and public_intensity[min_diff_idx] > max_prec_rel_int:
                        continue

                to_keep.append(annotation)

                # Add the reference spectrum to the public spectrum
                ref_spectrum = np.array(annotation.matched_spec)
                ref_mz = ref_spectrum[:, 0]
                ref_intensity = ref_spectrum[:, 1] / np.max(ref_spectrum[:, 1])

                if public_mz.size == 0:
                    public_mz = ref_mz
                    public_intensity = ref_intensity
                else:
                    # Add peaks
                    public_mz = np.concatenate([public_mz, ref_mz])
                    public_intensity = np.concatenate([public_intensity, ref_intensity])

            # Update annotations
            spec.annotation_ls = to_keep

    return ms1_spec_ls



import pandas as pd
import numpy as np


def align_feature_tables():
    """
    based on feature table from DDA MS2 analysis, align results from DDA MS1, Full scan MS1 0, 10, 20 eV
    """
    # read feature tables
    ms2_dda_df = pd.read_csv('../../../../data/nist/dda_ms2/aligned_feature_table.tsv', sep='\t', low_memory=False)
    ms1_dda_df = pd.read_csv('../../../../data/nist/dda_ms1/aligned_feature_table.tsv', sep='\t', low_memory=False)
    ms1_0ev_df = pd.read_csv('../../../../data/nist/fullscan_0ev/aligned_feature_table.tsv', sep='\t', low_memory=False)
    ms1_10ev_df = pd.read_csv('../../../../data/nist/fullscan_10ev/aligned_feature_table.tsv', sep='\t', low_memory=False)
    ms1_20ev_df = pd.read_csv('../../../../data/nist/fullscan_20ev/aligned_feature_table.tsv', sep='\t', low_memory=False)

    # align to DDA MS2
    ms2_dda_df['ms1_dda_pms2'] = [None] * ms2_dda_df.shape[0]
    ms2_dda_df['ms1_0ev_pms2'] = [None] * ms2_dda_df.shape[0]
    ms2_dda_df['ms1_10ev_pms2'] = [None] * ms2_dda_df.shape[0]
    ms2_dda_df['ms1_20ev_pms2'] = [None] * ms2_dda_df.shape[0]

    for df, col_name in zip([ms1_dda_df, ms1_0ev_df, ms1_10ev_df, ms1_20ev_df],
                            ['ms1_dda_pms2', 'ms1_0ev_pms2', 'ms1_10ev_pms2', 'ms1_20ev_pms2']):
        print(f'Aligning {col_name}...')

        for i, row in ms2_dda_df.iterrows():
            prec_mz = row['m/z']
            rt = row['RT']

            # find in feature table
            mask = ((df['m/z'] - prec_mz).abs() <= max(0.005, prec_mz * 1e-5)) & ((df['RT'] - rt).abs() <= 0.05)
            if sum(mask) > 0:
                pms2_list = df[mask]['pseudo_ms2'].tolist()
                # remove nan
                pms2_list = [x for x in pms2_list if not pd.isnull(x)]

                if len(pms2_list) > 0:
                    final_pms2_list = []
                    # convert to 2d numpy arrays
                    for pms2 in pms2_list:
                        pms2 = pms2.split(';')
                        pms2 = [x.split(' ') for x in pms2][:-1]
                        pms2 = [[float(x) for x in y] for y in pms2]
                        pms2 = np.array(pms2, dtype=np.float32)
                        final_pms2_list.append(pms2)
                    ms2_dda_df.at[i, col_name] = final_pms2_list

        print(f'{col_name}: {sum(ms2_dda_df[col_name].notnull())} features aligned')

    # save
    ms2_dda_df.to_pickle('data/aligned_feature_table_pms2.pkl')


from revcos_main import *
from matchms import Spectrum
def calc_revcos(reverse_cosine, pseudo_ms2, dda_ms2, prec_mz):

    # remove peaks with mz > precursor_mz
    pseudo_ms2 = pseudo_ms2[pseudo_ms2[:, 0] < prec_mz + 0.05]
    dda_ms2 = dda_ms2[dda_ms2[:, 0] < prec_mz + 0.05]

    # # remove peaks with relative intensity < 1%
    # pseudo_ms2 = pseudo_ms2[pseudo_ms2[:, 1] > pseudo_ms2[:, 1].max() * 0.01]
    # dda_ms2 = dda_ms2[dda_ms2[:, 1] > dda_ms2[:, 1].max() * 0.01]

    # sort by mz
    pseudo_ms2 = pseudo_ms2[np.argsort(pseudo_ms2[:, 0])]
    dda_ms2 = dda_ms2[np.argsort(dda_ms2[:, 0])]

    spectrum_1 = Spectrum(mz=dda_ms2[:, 0], intensities=np.sqrt(dda_ms2[:, 1]), metadata={"precursor_mz": prec_mz})
    spectrum_2 = Spectrum(mz=pseudo_ms2[:, 0], intensities=np.sqrt(pseudo_ms2[:, 1]), metadata={"precursor_mz": prec_mz})

    score, matched_peak, spec_usage_ref, spec_usage_qry, idx_ls_ref, idx_ls_qry = (
        reverse_cosine.pair(spectrum_1, spectrum_2))

    return (score, matched_peak, spec_usage_ref, spec_usage_qry,
            matched_peak / len(dda_ms2), matched_peak / len(pseudo_ms2))


def calc_spec_similarity():
    df = pd.read_pickle('data/aligned_feature_table_pms2.pkl')

    revcos_engine = ReverseCosine(tolerance=0.05)

    for col_name in ['ms1_dda_pms2', 'ms1_0ev_pms2', 'ms1_10ev_pms2', 'ms1_20ev_pms2']:
        df[f'{col_name}_similarity'] = [None] * df.shape[0]
        for i, row in df.iterrows():
            if pd.isnull(row['MS2']):
                continue

            if not isinstance(row[col_name], list):
                continue

            pms2_ls = row[col_name]

            # DDA MS2
            dda_ms2 = row['MS2']
            dda_ms2 = dda_ms2.split('|')
            dda_ms2 = [x.split(';') for x in dda_ms2]
            dda_ms2 = [[float(x) for x in y] for y in dda_ms2]
            dda_ms2 = np.array(dda_ms2, dtype=np.float32)

            score_ls = []
            matched_peak_ls = []
            spec_usage_qry_ls = []
            spec_usage_ref_ls = []
            match_ratio_qry_ls = []
            match_ratio_ref_ls = []
            for pms2 in pms2_ls:
                result = calc_revcos(revcos_engine, pms2, dda_ms2, row['m/z'])
                score_ls.append(result[0])
                matched_peak_ls.append(result[1])
                spec_usage_ref_ls.append(result[2])
                spec_usage_qry_ls.append(result[3])
                match_ratio_ref_ls.append(result[4])
                match_ratio_qry_ls.append(result[5])

                # peak scaling
                factor = np.exp(pms2[:, 0] / row['m/z'] * 10)
                pms2[:, 1] *= factor
                result = calc_revcos(revcos_engine, pms2, dda_ms2, row['m/z'])
                score_ls.append(result[0])
                matched_peak_ls.append(result[1])
                spec_usage_ref_ls.append(result[2])
                spec_usage_qry_ls.append(result[3])
                match_ratio_ref_ls.append(result[4])
                match_ratio_qry_ls.append(result[5])
            df.at[i, f'{col_name}_similarity'] = max(score_ls)

        print(f'{col_name}: {sum(df[f"{col_name}_similarity"].notnull())} similarities calculated')

    df.to_pickle('data/aligned_feature_table_pms2_revcos.pkl')


import matplotlib.pyplot as plt
def result_analysis():

    df = pd.read_pickle('data/aligned_feature_table_pms2_revcos.pkl')

    for col_name in ['ms1_dda_pms2', 'ms1_0ev_pms2', 'ms1_10ev_pms2', 'ms1_20ev_pms2']:
        sim_col_name = f'{col_name}_similarity'

        # histogram
        df[sim_col_name].hist(bins=50)
        plt.title(f'{col_name} similarity')
        plt.show()



if __name__ == '__main__':
    # align_feature_tables()

    # calc_spec_similarity()

    result_analysis()


# A module to read and process the raw MS data
# Classes are defined in order to handle the data

# Import modules
from pyteomics import mzml, mzxml
import numpy as np
import os
import matplotlib.pyplot as plt
import pandas as pd
from datetime import datetime
from time import time

from masscube.params import Params
from masscube_peak_dection import find_rois, cut_roi


class MSData:
    """
    A class that models a single file (mzML or mzXML) and
    processes the raw data.

    Attributes
    ----------------------------------------------------------
    scans: list, a list of Scan objects
    ms1_rt_seq: numpy array, retention times of all MS1 scans
    bpc_int: numpy array, intensity of the BPC
    rois: list, a list of ROI objects
    rois_mz_seq: numpy array, m/z of all ROIs
    params: Params object, a Params object that contains the parameters
    """

    def __init__(self):
        """
        Function to initiate MSData.
        ----------------------------------------------------------
        """

        self.scans = []  # A list of MS scans
        self.ms1_rt_seq = []  # Retention times of all MS1 scans
        self.bpc_int = []  # Intensity of the BPC
        self.rois = []  # A list of ROIs
        self.params = None  # A Params object
        self.file_name = None  # File name of the raw data without extension
        self.roi_mz_seq = None  # m/z of all ROIs
        self.roi_rt_seq = None  # Retention time of all ROIs
        self.start_time = None  # Start acquisition time of the raw data

    def read_raw_data(self, file_name, params, read_ms2=True, clean_ms2=False, centroid=True):
        """
        Function to read raw data to MS1 and MS2 (if available)
        (supported by pyteomics package).

        Parameters
        ----------------------------------------------------------
        file_name: str
            File name of raw MS data (mzML or mzXML).
        params: Params object
            A Params object that contains the parameters.
        """

        self.params = params

        if os.path.isfile(file_name):
            # get extension from file name
            ext = os.path.splitext(file_name)[1]

            if ext.lower() != ".mzml" and ext.lower() != ".mzxml":
                raise ValueError("Unsupported raw data format. Raw data must be in mzML or mzXML.")

            self.file_name = os.path.splitext(os.path.basename(file_name))[0]

            self.start_time = get_start_time(file_name)

            if ext.lower() == ".mzml":
                with mzml.MzML(file_name) as reader:
                    self.extract_scan_mzml(reader, int_tol=params.int_tol, read_ms2=read_ms2,
                                           clean_ms2=clean_ms2, centroid=centroid)
            elif ext.lower() == ".mzxml":
                with mzxml.MzXML(file_name) as reader:
                    self.extract_scan_mzxml(reader, int_tol=params.int_tol, read_ms2=read_ms2,
                                            clean_ms2=clean_ms2, centroid=centroid)
        else:
            print("File does not exist.")

    def extract_scan_mzml(self, spectra, int_tol=0, read_ms2=True, clean_ms2=False, centroid=True):
        """
        Function to extract all scans and convert them to Scan objects.

        Parameters
        ----------------------------------------------------------
        spectra: pyteomics object
            An iteratable object that contains all MS1 and MS2 scans.
        """

        # idx = 0  # Scan number
        self.ms1_idx = []  # MS1 scan index
        self.ms2_idx = []  # MS2 scan index

        rt_unit = spectra[0]['scanList']['scan'][0]['scan start time'].unit_info

        # Iterate over all scans
        for idx, spec in enumerate(spectra):
            # Get the retention time and convert to minute
            try:
                rt = spec['scanList']['scan'][0]['scan start time']
            except:
                rt = spec['scanList']['scan'][0]['scan time']

            if rt_unit == 'second':
                rt = rt / 60

            # Check if the retention time is within the range
            if self.params.rt_range[0] < rt < self.params.rt_range[1]:
                if spec['ms level'] == 1:
                    temp_scan = Scan(level=1, scan=idx, rt=rt)
                    mz_array = np.array(spec['m/z array'], dtype=np.float64)
                    int_array = np.array(spec['intensity array'], dtype=np.int64)
                    mz_array = mz_array[int_array > int_tol]
                    int_array = int_array[int_array > int_tol]

                    if centroid:
                        mz_array, int_array = _centroid(mz_array, int_array)

                    self.ms1_idx.append(idx)
                    self.ms1_rt_seq.append(rt)

                    if len(mz_array) == 0:
                        # update base peak chromatogram
                        self.bpc_int.append(0)
                        temp_scan.add_info_by_level(mz_seq=np.array([], dtype=np.float64),
                                                    int_seq=np.array([], dtype=np.int64))
                    else:
                        self.bpc_int.append(np.max(int_array))
                        temp_scan.add_info_by_level(mz_seq=mz_array, int_seq=int_array)


                elif spec['ms level'] == 2 and read_ms2:
                    temp_scan = Scan(level=2, scan=idx, rt=rt)
                    precursor_mz = spec['precursorList']['precursor'][0]['selectedIonList']['selectedIon'][0][
                        'selected ion m/z']
                    peaks = np.array([spec['m/z array'], spec['intensity array']], dtype=np.float64).T
                    temp_scan.add_info_by_level(precursor_mz=precursor_mz, peaks=peaks)
                    if clean_ms2:
                        _clean_ms2(temp_scan)
                    self.ms2_idx.append(idx)

                self.scans.append(temp_scan)
                # idx += 1

        self.ms1_rt_seq = np.array(self.ms1_rt_seq)

    def extract_scan_mzxml(self, spectra, int_tol, read_ms2=True, clean_ms2=False, centroid=True):
        """
        Function to extract all scans and convert them to Scan objects.

        Parameters
        ----------------------------------------------------------
        spectra: pyteomics object
            An iteratable object that contains all MS1 and MS2 scans.
        """

        # idx = 0  # Scan number
        self.ms1_idx = []  # MS1 scan index
        self.ms2_idx = []  # MS2 scan index

        rt_unit = spectra[0]["retentionTime"].unit_info

        # Iterate over all scans
        for idx, spec in enumerate(spectra):
            # Get the retention time and convert to minute
            rt = spec["retentionTime"]  # retention time of mzXML is in minute

            if rt_unit == 'second':
                rt = rt / 60

            # Check if the retention time is within the range
            if self.params.rt_range[0] < rt < self.params.rt_range[1]:
                if spec['msLevel'] == 1:
                    temp_scan = Scan(level=1, scan=idx, rt=rt)
                    mz_array = np.array(spec['m/z array'], dtype=np.float64)
                    int_array = np.array(spec['intensity array'], dtype=np.int64)

                    mz_array = mz_array[int_array > int_tol]
                    int_array = int_array[int_array > int_tol]

                    if centroid:
                        mz_array, int_array = _centroid(mz_array, int_array)

                    self.ms1_idx.append(idx)
                    self.ms1_rt_seq.append(rt)

                    if len(mz_array) == 0:
                        # update base peak chromatogram
                        self.bpc_int.append(0)
                        temp_scan.add_info_by_level(mz_seq=np.array([], dtype=np.float64),
                                                    int_seq=np.array([], dtype=np.int64))
                    else:
                        self.bpc_int.append(np.max(int_array))
                        temp_scan.add_info_by_level(mz_seq=mz_array, int_seq=int_array)

                elif spec['msLevel'] == 2 and read_ms2:
                    temp_scan = Scan(level=2, scan=idx, rt=rt)
                    precursor_mz = spec['precursorMz'][0]['precursorMz']
                    peaks = np.array([spec['m/z array'], spec['intensity array']], dtype=np.float64).T
                    temp_scan.add_info_by_level(precursor_mz=precursor_mz, peaks=peaks)
                    if clean_ms2:
                        _clean_ms2(temp_scan)
                    self.ms2_idx.append(idx)

                self.scans.append(temp_scan)
                # idx += 1

        self.ms1_rt_seq = np.array(self.ms1_rt_seq)

    def drop_ion_by_int(self):
        """
        Function to drop ions by intensity.

        Parameters
        ----------------------------------------------------------
        tol: int
            Intensity tolerance.
        """

        for idx in self.ms1_idx:
            self.scans[idx].mz_seq = self.scans[idx].mz_seq[self.scans[idx].int_seq > self.params.int_tol]
            self.scans[idx].int_seq = self.scans[idx].int_seq[self.scans[idx].int_seq > self.params.int_tol]

    def find_rois(self):
        """
        Function to find ROI in MS1 scans.

        Parameters
        ----------------------------------------------------------
        params: Params object
            A Params object that contains the parameters.
        """

        self.rois = find_rois(self)

        for roi in self.rois:
            roi.sum_roi(False, False)

        self.rois.sort(key=lambda x: x.mz)

    def cut_rois(self):
        """
        Function to cut ROI into smaller pieces.
        """

        self.rois = [cut_roi(r, int_tol=self.params.int_tol) for r in self.rois]
        tmp = []
        for roi in self.rois:
            tmp.extend(roi)

        self.rois = tmp

    def summarize_roi(self, cal_g_score=True, cal_a_score=True):
        """
        Function to process ROIs.

        Parameters
        ----------------------------------------------------------
        params: Params object
            A Params object that contains the parameters.
        """

        for roi in self.rois:
            roi.sum_roi(cal_g_score, cal_a_score)

        # sort rois by m/z
        self.rois.sort(key=lambda x: x.mz)

        # index the rois
        for idx in range(len(self.rois)):
            self.rois[idx].id = idx

        # extract mz and rt of all rois for further use (feature grouping)
        self.roi_mz_seq = np.array([roi.mz for roi in self.rois])
        self.roi_rt_seq = np.array([roi.rt for roi in self.rois])

        # allocate ms2 to rois
        self.allocate_ms2_to_rois()

        # find best ms2 for each roi
        for roi in self.rois:
            if len(roi.ms2_seq) > 0:
                roi.best_ms2 = find_best_ms2(roi.ms2_seq)

    def allocate_ms2_to_rois(self):
        for i in self.ms2_idx:
            if len(self.scans[i].peaks) == 0:
                continue
            idx = np.where(np.abs(self.roi_mz_seq - self.scans[i].precursor_mz) < 0.015)[0]
            matched_rois = []
            for j in idx:
                if self.rois[j].rt_seq[0] < self.scans[i].rt < self.rois[j].rt_seq[-1]:
                    matched_rois.append(self.rois[j])
            if len(matched_rois) > 0:
                # assign ms2 to the roi with the highest peak height
                matched_rois[np.argmax([roi.peak_height for roi in matched_rois])].ms2_seq.append(self.scans[i])

    def drop_rois_without_ms2(self):
        """
        Function to drop ROIs without MS2.
        """

        self.rois = [roi for roi in self.rois if len(roi.ms2_seq) > 0]

    def drop_rois_by_length(self, length=5):
        """
        Function to drop ROIs by length.
        """

        self.rois = [roi for roi in self.rois if roi.length >= length]

    def _discard_isotopes(self):
        """
        Function to discard isotopes.
        """

        self.rois = [roi for roi in self.rois if not roi.is_isotope]

        for idx in range(len(self.rois)):
            self.rois[idx].id = idx

    def plot_bpc(self, rt_range=None, label_name=False, output_dir=None):
        """
        Function to plot base peak chromatogram.

        Parameters
        ----------------------------------------------------------
        rt_range: list
            Retention time range [start, end]. The unit is minute.
        label_name: bool
            True: show the file name on the plot.
        output_dir: str
            Output directory of the plot. If specified, the plot will be saved to the directory.
            If None, the plot will be shown.
        """

        if rt_range is not None:
            x = [rt for rt in self.ms1_rt_seq if rt > rt_range[0] and rt < rt_range[1]]
            y = [intensity for rt, intensity in zip(self.ms1_rt_seq, self.bpc_int) if
                 rt > rt_range[0] and rt < rt_range[1]]

        else:
            x = self.ms1_rt_seq
            y = self.bpc_int

        plt.figure(figsize=(10, 3))
        plt.rcParams['font.size'] = 14
        plt.rcParams['font.family'] = 'Arial'
        plt.plot(x, y, linewidth=1, color="black")
        plt.xlabel("Retention Time (min)", fontsize=18, fontname='Arial')
        plt.ylabel("Intensity", fontsize=18, fontname='Arial')
        plt.xticks(fontsize=14, fontname='Arial')
        plt.yticks(fontsize=14, fontname='Arial')
        if label_name:
            plt.text(self.ms1_rt_seq[0], np.max(self.bpc_int) * 0.9, self.file_name, fontsize=12, fontname='Arial',
                     color="gray")

        if output_dir is not None:
            plt.savefig(output_dir, dpi=300, bbox_inches="tight")
            plt.close()
        else:
            plt.show()

    def output_single_file(self, user_defined_output_path=None):
        """
        Function to generate a report for rois in csv format.

        Parameters
        ----------------------------------------------------------
        user_defined_output_path: str
            User defined output path.
        """

        result = []

        for roi in self.rois:
            iso_dist = ""
            for i in range(len(roi.isotope_mz_seq)):
                iso_dist += str(np.round(roi.isotope_mz_seq[i], decimals=4)) + ";" + str(
                    np.round(roi.isotope_int_seq[i], decimals=0)) + "|"
            iso_dist = iso_dist[:-1]

            ms2 = ""
            if roi.best_ms2 is not None:
                for i in range(len(roi.best_ms2.peaks)):
                    ms2 += str(np.round(roi.best_ms2.peaks[i, 0], decimals=4)) + ";" + str(
                        np.round(roi.best_ms2.peaks[i, 1], decimals=0)) + "|"
                ms2 = ms2[:-1]

            temp = [roi.id, roi.mz.__round__(4), roi.rt.__round__(3), roi.length, roi.rt_seq[0],
                    roi.rt_seq[-1], roi.peak_area, roi.peak_height, roi.gaussian_similarity.__round__(2),
                    roi.noise_level.__round__(2), roi.asymmetry_factor.__round__(2), roi.charge_state, roi.is_isotope,
                    str(roi.isotope_id_seq)[1:-1], iso_dist,
                    roi.is_in_source_fragment, roi.isf_parent_roi_id, str(roi.isf_child_roi_id)[1:-1],
                    roi.adduct_type, roi.adduct_parent_roi_id, str(roi.adduct_child_roi_id)[1:-1],
                    ]

            temp.extend(
                [ms2, roi.annotation, roi.formula, roi.similarity, roi.matched_peak_number, roi.smiles, roi.inchikey])

            result.append(temp)

        # convert result to a pandas dataframe
        columns = ["ID", "m/z", "RT", "length", "RT_start", "RT_end", "peak_area", "peak_height",
                   "Gaussian_similarity", "noise_level", "asymmetry_factor", "charge", "is_isotope",
                   "isotope_IDs", "isotopes", "is_in_source_fragment", "ISF_parent_ID",
                   "ISF_child_ID", "adduct", "adduct_base_ID", "adduct_other_ID"]

        columns.extend(["MS2", "annotation", "formula", "similarity", "matched_peak_number", "SMILES", "InChIKey"])

        df = pd.DataFrame(result, columns=columns)

        # save the dataframe to csv file
        if user_defined_output_path:
            path = user_defined_output_path
        else:
            path = os.path.join(self.params.single_file_dir)
            if not os.path.exists(path):
                os.makedirs(path)
            path = os.path.join(self.params.single_file_dir, self.file_name + ".txt")
        df.to_csv(path, index=False, sep="\t")

    def get_eic_data(self, target_mz, target_rt=None, mz_tol=0.005, rt_tol=0.3, rt_range=None):
        """
        To get the EIC data of a target m/z.

        Parameters
        ----------
        target_mz: float
            Target m/z.
        mz_tol: float
            m/z tolerance.
        target_rt: float
            Target retention time.
        rt_tol: float
            Retention time tolerance.

        Returns
        -------
        eic_rt: numpy array
            Retention time of the EIC.
        eic_int: numpy array
            Intensity of the EIC.
        eic_mz: numpy array
            m/z of the EIC.
        eic_scan_idx: numpy array
            Scan index of the EIC.
        """

        eic_rt = []
        eic_int = []
        eic_mz = []
        eic_scan_idx = []

        if target_rt is not None:
            rt_range = [target_rt - rt_tol, target_rt + rt_tol]
        elif rt_range is None:
            rt_range = [0, np.inf]

        for i in self.ms1_idx:
            if self.scans[i].rt > rt_range[0] and self.scans[i].rt < rt_range[1]:
                mz_diff = np.abs(self.scans[i].mz_seq - target_mz)
                if len(mz_diff) > 0 and np.min(mz_diff) < mz_tol:
                    eic_rt.append(self.scans[i].rt)
                    eic_int.append(self.scans[i].int_seq[np.argmin(mz_diff)])
                    eic_mz.append(self.scans[i].mz_seq[np.argmin(mz_diff)])
                    eic_scan_idx.append(i)
                else:
                    eic_rt.append(self.scans[i].rt)
                    eic_int.append(0)
                    eic_mz.append(0)
                    eic_scan_idx.append(i)

            if self.scans[i].rt > rt_range[1]:
                break

        eic_rt = np.array(eic_rt)
        eic_int = np.array(eic_int)
        eic_mz = np.array(eic_mz)
        eic_scan_idx = np.array(eic_scan_idx)

        return eic_rt, eic_int, eic_mz, eic_scan_idx

    def plot_eic(self, target_mz, target_rt=None, mz_tol=0.005, rt_tol=0.3,
                 output=False, show_rt_line=True, return_eic_data=False):
        """
        Function to plot EIC of a target m/z.

        Parameters
        ----------
        target_mz: float
            Target m/z.
        target_rt: float
            Target retention time.
        mz_tol: float
            m/z tolerance.
        rt_tol: float
            Retention time tolerance.
        output: str
            Output file name. If not specified, the plot will be shown.
        show_rt_line: bool
            True: show the target retention time.
            False: do not show the target retention time.
        return_eic_data: bool
            True: return the EIC data.
            False: do not return the EIC data.
        """

        # get the eic data
        eic_rt, eic_int, _, _ = self.get_eic_data(target_mz, target_rt, mz_tol, rt_tol)

        plt.figure(figsize=(10, 3))
        plt.rcParams['font.size'] = 14
        plt.rcParams['font.family'] = 'Arial'
        plt.plot(eic_rt, eic_int, linewidth=1, color="black")
        plt.xlabel("Retention Time (min)", fontsize=18, fontname='Arial')
        plt.ylabel("Intensity", fontsize=18, fontname='Arial')
        plt.xticks(fontsize=14, fontname='Arial')
        plt.yticks(fontsize=14, fontname='Arial')
        if target_rt is not None and show_rt_line:
            plt.axvline(x=target_rt, color='b', linestyle='--', linewidth=1)

        if output:
            plt.savefig(output, dpi=300, bbox_inches="tight")
            plt.close()
        else:
            plt.show()

        if return_eic_data:
            return eic_rt, eic_int

    def find_ms2_by_mzrt(self, mz_target, rt_target, mz_tol=0.01, rt_tol=0.3, return_best=False):
        """
        Function to find MS2 scan by precursor m/z and retention time.

        Parameters
        ----------------------------------------------------------
        mz_target: float
            Precursor m/z.
        rt_target: float
            Retention time.
        mz_tol: float
            m/z tolerance.
        rt_tol: float
            Retention time tolerance.
        return_best: bool
            True: only return the best MS2 scan with the highest intensity.
            False: return all MS2 scans as a list.
        """

        matched_ms2 = []

        for id in self.ms2_idx:
            rt = self.scans[id].rt

            if rt < rt_target - rt_tol:
                continue

            mz = self.scans[id].precursor_mz

            if abs(mz - mz_target) < mz_tol and abs(rt - rt_target) < rt_tol:
                matched_ms2.append(self.scans[id])

            if rt > rt_target + rt_tol:
                break

        if return_best:
            if len(matched_ms2) > 1:
                total_ints = [np.sum(ms2.peaks[:, 1]) for ms2 in matched_ms2]
                return matched_ms2[np.argmax(total_ints)]
            elif len(matched_ms2) == 1:
                return matched_ms2[0]
            else:
                return None
        else:
            return matched_ms2

    def find_roi_by_mzrt(self, mz_target, rt_target=None, mz_tol=0.01, rt_tol=0.3):
        """
        Function to find roi by precursor m/z and retention time.

        Parameters
        ----------------------------------------------------------
        mz_target: float
            Precursor m/z.
        rt_target: float
            Retention time.
        mz_tol: float
            m/z tolerance.
        rt_tol: float
            Retention time tolerance.
        """

        if rt_target is None:
            tmp = np.abs(self.roi_mz_seq - mz_target) < mz_tol
            found_roi = [self.rois[i] for i in np.where(tmp)[0]]
        else:
            tmp1 = np.abs(self.roi_mz_seq - mz_target) < mz_tol
            tmp2 = np.abs(self.roi_rt_seq - rt_target) < rt_tol
            tmp = np.logical_and(tmp1, tmp2)
            found_roi = [self.rois[i] for i in np.where(tmp)[0]]

        return found_roi

    def find_ms1_scan_by_rt(self, rt_target):
        """
        Function to find a MS1 scan by retention time.

        Parameters
        ----------------------------------------------------------
        rt_target: float
            Retention time.
        """

        idx = np.argmin(np.abs(self.ms1_rt_seq - rt_target))
        return self.scans[self.ms1_idx[idx]]

    def correct_retention_time(self, f):
        """
        Function to correct retention time.

        Parameters
        ----------------------------------------------------------
        f: interp1d object
            A function to correct retention time.
        """

        all_rts = np.array([s.rt for s in self.scans])
        all_rts = f(all_rts)
        for i in range(len(self.scans)):
            self.scans[i].rt = all_rts[i]

    def plot_roi(self, roi_idx, mz_tol=0.005, rt_range=[0, np.inf], rt_window=None, output=False):
        """
        Function to plot EIC of a target m/z.
        """

        if rt_window is not None:
            rt_range = [self.rois[roi_idx].rt - rt_window, self.rois[roi_idx].rt + rt_window]

        # get the eic data
        eic_rt, eic_int, _, eic_scan_idx = self.get_eic_data(self.rois[roi_idx].mz, mz_tol=mz_tol, rt_range=rt_range)
        idx_start = np.where(eic_scan_idx == self.rois[roi_idx].scan_idx_seq[0])[0][0]
        idx_end = np.where(eic_scan_idx == self.rois[roi_idx].scan_idx_seq[-1])[0][0] + 1

        plt.figure(figsize=(7, 3))
        plt.rcParams['font.size'] = 14
        plt.rcParams['font.family'] = 'Arial'
        plt.plot(eic_rt, eic_int, linewidth=1, color="black")
        plt.fill_between(eic_rt[idx_start:idx_end], eic_int[idx_start:idx_end], color="black", alpha=0.4)
        plt.xlabel("Retention Time (min)", fontsize=18, fontname='Arial')
        plt.ylabel("Intensity", fontsize=18, fontname='Arial')
        plt.xticks(fontsize=14, fontname='Arial')
        plt.yticks(fontsize=14, fontname='Arial')

        if output:
            plt.savefig(output, dpi=300, bbox_inches="tight")
            plt.close()
            return None
        else:
            plt.show()
            return eic_rt[np.argmax(eic_int)], np.max(eic_int), eic_scan_idx[np.argmax(eic_int)]

    def plot_all_rois(self, output_path, mz_tol=0.01, rt_range=[0, np.inf], rt_window=None):
        """
        Function to plot EIC of all ROIs.
        """

        if output_path[-1] != "/":
            output_path += "/"

        for idx, roi in enumerate(self.rois):

            if rt_window is not None:
                rt_range = [roi.rt_seq[0] - rt_window, roi.rt_seq[-1] + rt_window]

            # get the eic data
            eic_rt, eic_int, _, eic_scan_idx = self.get_eic_data(roi.mz, mz_tol=mz_tol, rt_range=rt_range)
            idx_start = np.where(eic_scan_idx == roi.scan_idx_seq[0])[0][0]
            idx_end = np.where(eic_scan_idx == roi.scan_idx_seq[-1])[0][0] + 1

            plt.figure(figsize=(9, 3))
            plt.rcParams['font.size'] = 14
            plt.rcParams['font.family'] = 'Arial'
            plt.plot(eic_rt, eic_int, linewidth=0.5, color="black")
            plt.fill_between(eic_rt[idx_start:idx_end], eic_int[idx_start:idx_end], color="black", alpha=0.2)
            plt.axvline(x=roi.rt, color='b', linestyle='--', linewidth=1)
            plt.xlabel("Retention Time (min)", fontsize=18, fontname='Arial')
            plt.ylabel("Intensity", fontsize=18, fontname='Arial')
            plt.xticks(fontsize=14, fontname='Arial')
            plt.yticks(fontsize=14, fontname='Arial')
            plt.text(eic_rt[0], np.max(eic_int) * 0.95, "m/z = {:.4f}".format(roi.mz), fontsize=12, fontname='Arial')
            plt.text(eic_rt[0] + (eic_rt[-1] - eic_rt[0]) * 0.6, np.max(eic_int) * 0.95, self.file_name, fontsize=10,
                     fontname='Arial', color="gray")

            file_name = output_path + "roi{}_".format(idx) + str(roi.mz.__round__(4)) + ".png"

            plt.savefig(file_name, dpi=300, bbox_inches="tight")
            plt.close()


class Scan:
    """
    A class that represents a MS scan.
    A MS1 spectrum has properties including:
        scan number, retention time,
        m/z and intensities.
    A MS2 spectrum has properties including:
        scan number, retention time,
        precursor m/z, product m/z and intensities.
    """

    def __init__(self, level=None, scan=None, rt=None):
        """
        Function to initiate MS1Scan by precursor mz,
        retention time.

        Parameters
        ----------------------------------------------------------
        level: int
            Level of MS scan.
        scan: int
            Scan number.
        rt: float
            Retention time.
        """

        self.level = level
        self.scan = scan
        self.rt = rt

        # for MS1 scans:
        self.mz_seq = None
        self.int_seq = None

        # for MS2 scans:
        self.precursor_mz = None
        self.peaks = None

    def add_info_by_level(self, **kwargs):
        """
        Function to add scan information by level.
        """

        if self.level == 1:
            self.mz_seq = kwargs['mz_seq']
            self.int_seq = np.int64(kwargs['int_seq'])

        elif self.level == 2:
            self.precursor_mz = kwargs['precursor_mz']
            self.peaks = kwargs['peaks']

    def show_scan_info(self):
        """
        Function to print a scan's information.

        Parameters
        ----------------------------------------------------------
        scan: MS1Scan or MS2Scan object
            A MS1Scan or MS2Scan object.
        """

        print("Scan number: " + str(self.scan))
        print("Retention time: " + str(self.rt))

        if self.level == 1:
            print("m/z: " + str(np.around(self.mz_seq, decimals=4)))
            print("Intensity: " + str(np.around(self.int_seq, decimals=0)))

        elif self.level == 2:
            # keep 4 decimal places for m/z and 0 decimal place for intensity
            print("Precursor m/z: " + str(np.round(self.precursor_mz, decimals=4)))
            print(self.peaks)

    def plot_scan(self, mz_range=None, return_data=False):
        """
        Function to plot a scan.

        Parameters
        ----------------------------------------------------------
        """

        if self.level == 1:
            x = self.mz_seq
            y = self.int_seq
        elif self.level == 2:
            x = self.peaks[:, 0]
            y = self.peaks[:, 1]

        if mz_range is None:
            mz_range = [np.min(x) - 10, np.max(x) + 10]
        else:
            y = y[np.logical_and(x > mz_range[0], x < mz_range[1])]
            x = x[np.logical_and(x > mz_range[0], x < mz_range[1])]

        max_int = np.max(y)
        x_left = x[0] - 1
        x_right = x[-1] + 1
        plt.figure(figsize=(10, 3))
        plt.rcParams['font.size'] = 14
        plt.rcParams['font.family'] = 'Arial'
        plt.ylim(0, max_int * 1.2)
        plt.xlim(x_left, x_right)
        plt.vlines(x=x, ymin=0, ymax=y, color="black", linewidth=1.5)
        plt.hlines(y=0, xmin=mz_range[0], xmax=mz_range[1], color="black", linewidth=1.5)
        plt.xlabel("m/z, Dalton", fontsize=18, fontname='Arial')
        plt.ylabel("Intensity", fontsize=18, fontname='Arial')
        plt.xticks(fontsize=14, fontname='Arial')
        plt.yticks(fontsize=14, fontname='Arial')

        if self.level == 2:
            plt.text(x_left + 1, max_int * 1.1, "Precursor m/z = {:.4f}".format(self.precursor_mz), fontsize=11,
                     fontname='Arial')

        plt.text(x_left + 1 + (x_right - x_left) * 0.3, max_int * 1.1, "RT = {:.3f} min".format(self.rt), fontsize=11,
                 fontname='Arial')

        plt.show()

        if return_data:
            return x, y


def _clean_ms2(ms2, offset=2, int_drop_ratio=0.01):
    """
    A function to clean MS/MS by
    1. Drop ions with m/z > (precursor_mz - offset)
    2. Drop ions with intensity < 1% of the base peak intensity

    Parameters
    ----------
    ms2: Scan object
        A MS2 scan object.
    offset: float
        m/z offset for dropping ions.
    """

    if ms2.peaks.shape[0] > 0:
        ms2.peaks = ms2.peaks[ms2.peaks[:, 0] < ms2.precursor_mz - offset]
    if ms2.peaks.shape[0] > 0:
        ms2.peaks = ms2.peaks[ms2.peaks[:, 1] > int_drop_ratio * np.max(ms2.peaks[:, 1])]


def _centroid(mz_seq, int_seq, mz_tol=0.005):
    """
    Function to centroid the m/z and intensity sequences.

    Parameters
    ----------
    mz_seq: numpy array or list
        m/z sequence.
    int_seq: numpy array or list
        Intensity sequence.
    mz_tol: float
        m/z tolerance for centroiding. Default is 0.005.
    """

    diff = np.diff(mz_seq)
    tmp = np.where(diff < mz_tol)[0]

    if len(tmp) == 0:
        return mz_seq, int_seq

    mz_seq = list(mz_seq)
    int_seq = list(int_seq)
    for i in tmp[::-1]:
        mz_seq[i] = (mz_seq[i] * int_seq[i] + mz_seq[i + 1] * int_seq[i + 1]) / (int_seq[i] + int_seq[i + 1])
        int_seq[i] += int_seq[i + 1]
        mz_seq.pop(i + 1)
        int_seq.pop(i + 1)

    return np.array(mz_seq), int_seq


def read_raw_file_to_obj(file_name, params=None, int_tol=1000, centroid=True, read_ms2=True,
                         clean_ms2=False, print_summary=False):
    """
    Read a raw file to a MSData object.
    It's a useful function for data visualization or brief data analysis.

    Parameters
    ----------
    file_name : str
        The file name of the raw file.
    params : Params object
        A Params object that contains the parameters.
    int_tol : float
        Intensity tolerance for dropping ions.
    centroid : bool
        True: centroid the raw data.
        False: do not centroid the raw data.
    print_summary : bool
        True: print the summary of the raw data.
        False: do not print the summary of the raw data.

    Returns
    -------
    d : MSData object
        A MSData object.
    """

    # create a MSData object
    d = MSData()

    # read raw data
    if params is None:
        params = Params()
        params.int_tol = int_tol

    d.read_raw_data(file_name, params=params, read_ms2=read_ms2, clean_ms2=clean_ms2, centroid=centroid)

    if print_summary:
        print("Number of MS1 scans: " + str(len(d.ms1_idx)), "Number of MS2 scans: " + str(len(d.ms2_idx)))

    return d


def find_best_ms2(ms2_seq):
    """
    Function to find the best MS2 spectrum for a list of MS2 spectra.
    """

    if len(ms2_seq) > 0:
        total_ints = [np.sum(ms2.peaks[:, 1]) for ms2 in ms2_seq]
        if np.max(total_ints) == 0:
            return None
        else:
            return ms2_seq[max(range(len(total_ints)), key=total_ints.__getitem__)]
    else:
        return None


def get_start_time(file_name):
    """
    Function to get the start time of the raw data.

    Parameters
    ----------
    file_name : str
        Absolute path of the raw data.
    """

    with open(file_name, "rb") as f:
        for l in f:
            l = str(l)
            if "startTimeStamp" in str(l):
                t = l.split("startTimeStamp")[1].split('"')[1]
                return datetime.strptime(t, "%Y-%m-%dT%H:%M:%SZ")


import matplotlib.pyplot as plt
import numpy as np

# Data
omnivore = [4032568.0, 5721773.0]
vegan = [827748.0, 2833706.0]

# Set up the plot style
plt.rcParams['font.family'] = 'Arial'

# Define color scheme
group_colors = {'Omnivore': '#56648a', 'Vegan': '#facaa9'}

# Create the scatter plot
fig, ax = plt.subplots(figsize=(1.8, 1.8))

# Define groups and their positions
groups = ['Omnivore', 'Vegan']
positions = {group: i for i, group in enumerate(groups)}

# Plot scatter points
for i, (group, data) in enumerate(zip(groups, [omnivore, vegan])):
    x = np.random.normal(i, 0.05, len(data))
    ax.scatter(x, data, alpha=0.9, s=25, color=group_colors[group], edgecolors='none', zorder=2)

# Add group means
for i, (group, data) in enumerate(zip(groups, [omnivore, vegan])):
    mean = np.mean(data)
    ax.plot([i-0.2, i+0.2], [mean, mean], color='0.5', linewidth=1.5, zorder=1)

# Customize the plot
ax.set_ylabel('Peak height', fontsize=10, fontname='Arial', labelpad=3.5, color='0.2')
ax.set_ylim(0, 7e6)

# Customize ticks and remove top and right spines
ax.set_xticks([0, 1])
ax.set_xticklabels(groups)
ax.tick_params(axis='x', which='major', labelsize=10, pad=4, length=1.2, colors='0.2')
ax.tick_params(axis='y', which='major', labelsize=10, pad=2.5, length=2, colors='0.2')
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['bottom'].set_color('0.5')
ax.spines['left'].set_color('0.5')

# Set x-axis and y-axis label font to Arial and color to 0.2
for label in ax.get_xticklabels() + ax.get_yticklabels():
    label.set_fontname('Arial')
    label.set_color('0.2')

plt.tight_layout()

# Save as svg
plt.savefig('data/scatter_plot_omnivore_vs_vegan.svg', transparent=True, bbox_inches='tight')
plt.show()